# coding=utf-8
from __future__ import division
from . import ph, np, TestData, approx


def test_phasic_estimation():
    # %%
    FSAMP = 2048

    # %%
    # EDA
    eda = ph.EvenlySignal(TestData.eda(), sampling_freq=FSAMP, signal_nature='EDA', start_time=0)

    # preprocessing
    # downsampling
    eda = eda.resample(8)

    # filter
    eda = ph.IIRFilter(fp=0.8, fs=1.1)(eda)

    # %%
    # estimate driver
    driver = ph.DriverEstim(t1=0.75, t2=2.5)(eda)
    assert (int(np.mean(driver) * 10000) == 18286)
    assert (isinstance(driver, ph.EvenlySignal))

    # %%
    phasic, tonic, driver_no_peak = ph.PhasicEstim(delta=0.1)(driver)
    assert np.mean(phasic) == approx(.0656, abs=0.0005)
    assert (isinstance(phasic, ph.EvenlySignal))



# coding=utf-8
from abc import abstractmethod as _abstract, ABCMeta as _ABCMeta
from copy import copy as _cpy
import numpy as _np
from .Utility import PhUI as _PhUI
from .BaseAlgorithm import Algorithm as _Algorithm, Cache
from .Signal import EvenlySignal as _EvenlySignal, UnevenlySignal as _UnevenlySignal

__author__ = 'AleB'


class Segment(object):
    """
    Base Segment, a time begin-end pair with a reference to the base signal and a name.
    """

    def __init__(self, begin, end, label=None, signal=None):
        """
        Creates a base Window
        @param begin: Begin sample index
        @param end: End sample index
        """
        self._begin = begin
        self._end = end
        self._label = label
        self._signal = signal

    def get_begin_time(self):
        return self._begin

    def get_end_time(self):
        return self._end

    def get_begin(self):
        return self._signal.get_idx(self.get_begin_time())

    def get_end(self):
        return self._signal.get_idx(self.get_end_time()) if self.get_end_time() is not None else None

    def get_duration(self):
        return (self.get_end_time() - self.get_begin_time()) if self.get_end_time() is not None else None

    def get_label(self):
        return self._label

    def is_empty(self):
        return self._signal is None or self.get_begin_time() >= len(self._signal.get_end_time())

    def __call__(self, data=None):
        if data is None:
            data = self._signal
        return data.segment_time(self.get_begin_time(), self.get_end_time())

    def __repr__(self):
        return '[%s:%s' % (str(self.get_begin_time()), str(self.get_end_time())) + (
            ":%s]" % self._label if self._label is not None else "]")


class SegmentsGenerator(_Algorithm):
    """
    Base and abstract class for the windows computation.
    """
    __metaclass__ = _ABCMeta

    @_abstract
    def __init__(self, **kwargs):
        super(SegmentsGenerator, self).__init__(**kwargs)
        self._signal = None

    @_abstract
    def next_segment(self):
        """
        Executes a segmentation step.
        @raise StopIteration: End of the iteration
        """
        raise StopIteration()

    @_abstract
    def init_segmentation(self):
        """
        Executes a segmentation step.
        @raise StopIteration: End of the iteration
        """
        raise NotImplementedError()

    # Algorithm Override, no cache
    def __call__(self, data=None):
        assert data is not None or self._signal is not None, "No signal specified for " + self.__class__.__name__
        return self.run(data if data is not None else self._signal, self._params, use_cache=False)

    def __iter__(self):
        return SegmentationIterator(self)

    @classmethod
    def algorithm(cls, data, params):
        o = cls(**params)
        o._signal = data
        return o

    @classmethod
    def is_nature_supported(cls, data):
        return isinstance(data, _EvenlySignal)

    @classmethod
    def get_used_params(cls):
        return []

    def __repr__(self):
        if self._signal is not None:
            return super(SegmentsGenerator, self).__repr__() + " over\n" + str(self._signal)
        else:
            return super(SegmentsGenerator, self).__repr__()

    #REWRITE the run method for compatibility with MultiEvenly signals
    #Segments should not work on each channel
    #But output a sehments with multiple channels
    #to be processed by the algorithms
    
    def run(cls, data, params=None, use_cache=False, **kwargs):
        """
        Gets the data from the cache or calculates, caches and returns it.
        @param data: Source data
        @type data: TimeSeries
        @param params: Parameters for the calculator
        @type params: dict
        @param use_cache: Whether to use the cache memory or not
        @type use_cache: bool
        @return: The value of the feature.
        """
        if type(params) is dict:
            kwargs.update(params)
        if not isinstance(data.get_values(), _np.ndarray):
            _PhUI.w("The data must be a Signal (see class EvenlySignal and UnevenlySignal).")
            use_cache = False
        if use_cache is True:
            Cache.cache_check(data)
            # noinspection PyTypeChecker
            return Cache.run_cached(data, cls, kwargs)
        else:        
            return cls.algorithm(data, kwargs)            
                    
class SegmentationIterator(object):
    """
    A generic iterator that is called from each WindowGenerator from the __iter__ method.
    """

    def __init__(self, win):
        assert isinstance(win, SegmentsGenerator)
        self._win = _cpy(win)
        self._win.init_segmentation()

    def __next__(self):
        return self._win.next_segment()

    # Python 2 & users compatibility
    def next(self):
        return self.__next__()


class SegmentationError(Exception):
    """
    Generic Segmentation error.
    """
    pass



# coding=utf-8
from pyphysio.BaseAlgorithm import Algorithm

__author__ = 'AleB'


class Tool(Algorithm):
    """
    Algorithms that take as input a signal and return one or more np.array
    """

    @classmethod
    def get_signal_type(cls):
        return None

    @classmethod
    def is_compatible(cls, signal):
        return cls.get_signal_type() is None or signal.get_signal_nature() in cls.get_signal_type()



# coding=utf-8
from __future__ import division
import numpy as _np
from scipy.signal import welch as _welch, periodogram as _periodogram, freqz as _freqz
import scipy.optimize as _opt
import pycwt.wavelet as wave
from scipy import linalg as _linalg

import itertools as _itertools
from ..BaseTool import Tool as _Tool
from ..Signal import UnevenlySignal as _UnevenlySignal, EvenlySignal as _EvenlySignal


class Diff(_Tool):
    """
    Computes the differences between adjacent samples.

    Optional parameters
    -------------------
    degree : int, >0, default = 1
        Sample interval to compute the differences
    
    Returns
    -------
    signal : 
        Differences signal. 

    """

    def __init__(self, degree=1):
        assert degree > 0, "The degree value should be positive"
        _Tool.__init__(self, degree=degree)

    @classmethod
    def algorithm(cls, signal, params):
        """
        Calculates the differences between consecutive values
        """
        degree = params['degree']

        sig_1 = signal[:-degree]
        sig_2 = signal[degree:]

        out = _EvenlySignal(values=sig_2 - sig_1,
                            sampling_freq=signal.get_sampling_freq(),
                            signal_type=signal.get_signal_type(),
                            start_time=signal.get_start_time() + degree / signal.get_sampling_freq())

        return out
class PeakDetection(_Tool):
    """
    Estimate the maxima and the minima in the signal (in particular for periodic signals).

    Parameters
    ----------
    delta : float or list
        Threshold for the detection of the peaks. If it is a list it must have the same length of the signal.
        
    Optional parameters
    -------------------
    refractory : float, >=0, default = 0
        Seconds to skip after a detected paek to look for new peaks.
    start_max : boolean, default = True
        Whether to start looking for a maximum or (False) for a minimum.

    Returns
    -------
    maxp : numpy.array
        Array containing indexes of the maxima
    minp : numpy.array
        Array containing indexes of the minima
    maxv : numpy.array
        Array containing values of the maxima
    minv : numpy.array
        Array containing values of the minima
    """

    def __init__(self, delta, refractory=0, start_max=True):
        delta = _np.array(delta)
        assert delta.ndim <= 1, "Delta value should be 1 or 0-dimensional"
        assert delta.all() > 0, "Delta value/s should be positive"
        assert refractory >= 0, "Refractory value should be non negative"
        _Tool.__init__(self, delta=delta, refractory=refractory, start_max=start_max)

    @classmethod
    def algorithm(cls, signal, params):
        refractory = params['refractory']
        if refractory == 0:  # if 0 then do not skip samples
            refractory = 1
        else:  # else transform the refractory from seconds to samples
            refractory = refractory * signal.get_sampling_freq()
        look_for_max = params['start_max']
        delta = params['delta']

        minp = []
        maxp = []

        minv = []
        maxv = []

        scalar = delta.ndim == 0
        if scalar:
            d = delta

        if len(signal) < 1:
            cls.warn("Empty signal (len < 1), returning empty.")
        elif not scalar and len(delta) != len(signal):
            cls.error("delta vector's length differs from signal's one, returning empty.")
        else:
            mn_pos_candidate = mx_pos_candidate = 0
            mn_candidate = mx_candidate = signal[0]

            i_activation_min = 0
            i_activation_max = 0

            for i in range(1, len(signal)):
                sample = signal[i]
                if not scalar:
                    d = delta[i]

                if sample > mx_candidate:
                    mx_candidate = sample
                    mx_pos_candidate = i
                if sample < mn_candidate:
                    mn_candidate = sample
                    mn_pos_candidate = i

                if look_for_max:
                    if i >= i_activation_max and sample < mx_candidate - d:  # new max
                        maxp.append(mx_pos_candidate)
                        maxv.append(mx_candidate)
                        i_activation_max = i + refractory

                        mn_candidate = sample
                        mn_pos_candidate = i

                        look_for_max = False
                else:
                    if i >= i_activation_min and sample > mn_candidate + d:  # new min
                        minp.append(mn_pos_candidate)
                        minv.append(mn_candidate)
                        i_activation_min = i + refractory

                        mx_candidate = sample
                        mx_pos_candidate = i

                        look_for_max = True

        return _np.array(maxp), _np.array(minp), _np.array(maxv), _np.array(minv)


class PeakSelection(_Tool):
    """
    Identify the start and the end indexes of each peak in the signal, using derivatives.

    Parameters
    ----------
    indices : array, >=0
        Array containing indexes (first column) and values (second column) of the maxima
    win_pre : float, >0
        Duration (in seconds) of interval before the peak that is considered to find the start of the peak
    win_post : float, >0
        Duration (in seconds) of interval after the peak that is considered to find the end of the peak
    
    Returns
    -------
    starts : array
        Array containing start indexes
    ends : array
        Array containing end indexes
    """

    def __init__(self, indices, win_pre, win_post):
        indices = _np.array(indices)
        assert indices.ndim < 2, "Parameter indices has to be 1 or 0-dimensional"
        assert indices.all() >= 0, "Parameter indices contains negative values"
        assert win_pre > 0, "Window pre peak value should be positive"
        assert win_post > 0, "Window post peak value should be positive"
        _Tool.__init__(self, indices=indices, win_pre=win_pre, win_post=win_post)

    @classmethod
    def algorithm(cls, signal, params):
        i_peaks = params['indices']
        i_pre_max = int(params['win_pre'] * signal.get_sampling_freq())
        i_post_max = int(params['win_post'] * signal.get_sampling_freq())

        ZERO = 0.01

        i_start = _np.empty(len(i_peaks), int)
        i_stop = _np.empty(len(i_peaks), int)

        signal_dt = Diff()(signal)
        for i in range(len(i_peaks)):
            i_pk = int(i_peaks[i])

            if i_pk < i_pre_max:
                i_st = 0
                i_sp = i_pk + i_post_max
            elif i_pk >= len(signal_dt) - i_post_max:
                i_st = i_pk - i_pre_max
                i_sp = len(signal_dt) - 1
            else:
                i_st = i_pk - i_pre_max
                i_sp = i_pk + i_post_max

            # find START
            signal_dt_pre = signal_dt[i_st:i_pk]
            i_pre = len(signal_dt_pre) - 1

            # OR below is to allow small fluctuations (?)

            while i_pre > 0 and (signal_dt_pre[i_pre] > 0 or abs(signal_dt_pre[i_pre]) <= ZERO):
                i_pre -= 1

            i_start[i] = i_st + i_pre + 1

            # find STOP
            signal_dt_post = signal_dt[i_pk: i_sp]
            i_post = 1

            # OR below is to allow small fluctuations (?)
            while i_post < len(signal_dt_post) - 1 and (
                            signal_dt_post[i_post] < 0 or abs(signal_dt_post[i_post]) <= ZERO):
                i_post += 1

            i_stop[i] = i_pk + i_post

        return i_start, i_stop


class SignalRange(_Tool):
    """
    Estimate the local range of the signal by sliding windowing

    Parameters
    ----------
    win_len : float, >0
        Length of the window  in seconds
    win_step : float, >0
        Shift to start the next window in seconds

    Optional parameters
    -------------------    
    smooth : boolean, default=True
        Whether to convolve the result with a gaussian window

    Returns
    -------
    deltas : numpy.array
        Local range of the signal
    """

    def __init__(self, win_len, win_step, smooth=True):
        assert win_len > 0, "Window length should be positive"
        assert win_step > 0, "Window step should be positive"
        _Tool.__init__(self, win_len=win_len, win_step=win_step, smooth=smooth)

    @classmethod
    def algorithm(cls, signal, params):
        win_len = params['win_len']
        win_step = params['win_step']
        smooth = params['smooth']

        fsamp = signal.get_sampling_freq()
        idx_len = int(win_len * fsamp)
        idx_step = int(win_step * fsamp)

        if len(signal) < idx_len:
            cls.warn("Input signal is shorter than the window length.")
            return _np.max(signal) - _np.min(signal)
        else:
            windows = _np.arange(0, len(signal) - idx_len + 1, idx_step)
            deltas = _np.zeros(len(signal))

            curr_delta = 0
            for start in windows:
                portion_curr = signal[start: start + idx_len]
                curr_delta = _np.max(portion_curr) - _np.min(portion_curr)
                deltas[start:start + idx_len] = curr_delta

            deltas[windows[-1] + idx_len:] = curr_delta

            if smooth:
                win_len = int(win_len*2*fsamp)
                deltas = _np.convolve(deltas, _np.ones(win_len)/win_len, mode='same')

            return deltas


class PSD(_Tool):
    """
    Estimate the power spectral density (PSD) of the signal.

    Parameters
    ----------
    method : str
        Method to estimate the PSD. Available methods: 'welch', 'fft', 'ar'
        
    Optional parameters
    -------------------
    
    nfft : int, >0, default=2048
        Number of samples of the PSD
    window : str, default = 'hamming'
        Type of window
    min_order : int, >0, default=18
        Minimum order of the model to be tested for psd_method='ar'
    max_order : int, >0, default=25
        Maximum order of the model to be tested for psd_method='ar'
    normalize : boolean, default = True
        Whether to normalize the PSD
    remove_mean : boolean, default = True
        Whether to remove the mean from the signal before estimating the PSD
    
    Returns
    -------
    freq : numpy.array
        Frequencies
    psd : numpy.array
        Power Spectrum Density
    """

    def __init__(self, method, nfft=2048, window='hamming', min_order=10, max_order=30, normalize=False,
                 remove_mean=True, **kwargs):
        
        _method_list = ['welch', 'fft', 'ar']
        _window_list = ['hamming', 'blackman', 'hanning', 'bartlett', 'none']

        assert method in _method_list, "Parameter method should be in " + _method_list.__repr__()
        assert nfft > 0, "nfft value should be positive"
        assert window in _window_list, "Parameter window type should be in " + _window_list.__repr__()
        if method == "ar":
            assert min_order > 0, "Minimum order for the AR method should be positive"
            assert max_order > 0, "Maximum order for the AR method should be positive"
        
        _Tool.__init__(self, method=method, nfft=nfft, window=window, min_order=min_order,
                       max_order=max_order, normalize=normalize, remove_mean=remove_mean, **kwargs)

    # TODO (Feature - Issue #15): consider point below:
    # A density spectrum considers the amplitudes per unit frequency.
    # Density spectra are used to compare spectra with different frequency resolution as the
    # magnitudes are not influenced by the resolution because it is per Hertz. The amplitude
    # spectra on the other hand depend on the chosen frequency resolution.

    @classmethod
    def algorithm(cls, signal, params):
        method = params['method']
        nfft = params['nfft'] if "nfft" in params else None
        window = params['window']
        normalize = params['normalize']
        remove_mean = params['remove_mean']

        assert isinstance(signal, _EvenlySignal), "The PSD can be computed on EvenlySignals only. Consider interpolating the signal: signal.resample(fsamp)"

        fsamp = signal.get_sampling_freq()

        if remove_mean:
            signal = signal - _np.mean(signal)

        if method == 'fft':
            freqs, psd = _periodogram(signal, fs=fsamp, window = window, nfft=nfft, return_onesided=True)

        elif method == 'welch':
            freqs, psd = _welch(signal, fsamp, window=window, return_onesided=True, nfft=nfft)

        elif method == 'ar':
            cls.warn("Using AR method: results might not be comparable with other methods")
            #methods derived from: https://github.com/mpastell/pyageng
            def autocorr(x, lag=30):
                c = _np.correlate(x, x, 'full')
                mid = len(c)//2
                acov = c[mid:mid+lag]
                acor = acov/acov[0]
                return(acor)
                
            def aryw(x, order=30):
                x = x - _np.mean(x)
                ac = autocorr(x, order+1)
                R = _linalg.toeplitz(ac[:order])
                r = ac[1:order+1]
                params = _np.linalg.inv(R).dot(r)
                return(params)
                
            def AIC_yule(signal, order):
                #this is from library spectrum: https://github.com/cokelaer/spectrum
                N = len(signal)
                assert N>=order, "The number of samples in the signal should be >= to the model order"
                
                C = _np.correlate(signal, signal, mode='full')/N
                r = C[N-1:]
                
                T0  = r[0]
                T = r[1:]
                
                A = _np.zeros(order, dtype=float)
                P = T0
                
                for k in range(0, order):
                    save = T[k]
                    if k == 0:
                        temp = -save / P
                    else:
                        for j in range(0, k):
                            save = save + A[j] * T[k-j-1]
                        temp = -save / P
                    
                    P = P * (1. - temp**2.)
                    A[k] = temp
                
                    khalf = (k+1)//2
                    for j in range(0, khalf):
                        kj = k-j-1
                        save = A[j]
                        A[j] = save + temp * A[kj]
                        if j != kj:
                            A[kj] += temp*save
                
                res = N * _np.log(P) + 2*(order + 1)
                return(res)
            
            min_order = params['min_order']
            max_order = params['max_order']

            if len(signal) <= max_order:
                cls.warn("Input signal too short: try another 'method', a lower 'max_order', or a longer signal")
                return [], []

            orders = _np.arange(min_order, max_order + 1)
            aics = [AIC_yule(signal, x) for x in orders]
            best_order = orders[_np.argmin(aics)]

            params = aryw(signal, best_order)
            a = _np.concatenate([_np.ones(1), -params])
            w, P = _freqz(1, a, whole = False, worN = nfft)
            
            psd = 2*_np.abs(P)/fsamp
            
        else:
            cls.warn('Method not understood, using welch.')
            bands_w, psd = _welch(signal, fsamp, nfft=nfft, scaling = 'spectrum')

        freqs = _np.linspace(start=0, stop=fsamp / 2, num=len(psd))

        # NORMALIZE
        if normalize:
            psd /= _np.sum(psd)
        return freqs, psd


class Wavelet(_Tool):
    """
    TODO
    """
    def __init__(self, detrend=True, mother = None, **kwargs):
        mother = wave.Morlet(6) if mother is None else mother
        _Tool.__init__(self, detrend = detrend, mother = mother, **kwargs)
        

    @classmethod
    def algorithm(self, signal, params):
        t = signal.get_times()
        t0 = signal.get_start_time()
        dt = 1/signal.get_sampling_freq()
        
        detrend = params['detrend']
        if detrend:
            #% detrend
            p = _np.polyfit(t - t0, signal, 1)
            signal = signal - _np.polyval(p, t - t0)
            
        #% wavelet
        mother = params['mother']
        w, scales, freqs, coi, fft, fftfreqs = wave.cwt(signal, dt, wavelet=mother)
        
        power = (_np.abs(w)) ** 2
        power /= scales[:, None]
        return freqs, power

class Maxima(_Tool):
    """
    Find all local maxima in the signal

    Parameters
    ----------
    win_len : float, >0
        Length of window in seconds (method = 'windowing')
    win_step : float, >0
        Shift of the window to start the next window in seconds (method = 'windowing')
    method : str
        Method to detect the maxima. Available methods: 'complete' or 'windowing'. 'complete' finds all the local
         maxima, 'windowing' uses a runnning window to find the global maxima in each window.
    
    Optional parameters
    -------------------
    refractory : float, >0, default=0
        Seconds to skip after a detected maximum to look for new maxima, when method = 'complete'. 

    Returns
    -------
    idx_maxs : array
        Array containing indexes of the maxima
    val_maxs : array
        Array containing values of the maxima
    """

    def __init__(self, method='complete', refractory=0, win_len=None, win_step=None):
        assert method in ['complete', 'windowing'], "Method not valid"
        assert refractory >= 0, "Refractory time value should be positive (or 0 to deactivate)"
        
        if method == 'windowing':
            assert win_len > 0, "Window length should be positive"
            assert win_step > 0, "Window step should be positive"
            _Tool.__init__(self, method=method, refractory=refractory, win_len=win_len, win_step=win_step)
        elif method == 'complete':
            _Tool.__init__(self, method=method, refractory=refractory)
        
        

    @classmethod
    def algorithm(cls, signal, params):
        method = params['method']
        if method == 'complete':
            refractory = params['refractory']
            if refractory == 0:
                refractory = 1
            else:
                refractory = refractory * signal.get_sampling_freq()
            idx_maxs = []
            prev = signal[0]
            k = 1
            while k < len(signal) - 1 - refractory:
                curr = signal[k]
                nxt = signal[k + 1]
                if (curr >= prev) and (curr >= nxt):
                    idx_maxs.append(k)
                    prev = signal[k + 1 + refractory]
                    k = k + 2 + refractory
                else:  # continue
                    prev = signal[k]
                    k += 1
            idx_maxs = _np.array(idx_maxs).astype(int)
            maxs = signal[idx_maxs]
            return idx_maxs, maxs
        elif method == 'windowing':
            fsamp = signal.get_sampling_freq()
            winlen = int(params['win_len'] * fsamp)
            winstep = int(params['win_step'] * fsamp)

            # TODO (Andrea): check that winlen > 2
            # TODO (Andrea): check that winstep >= 1

            idx_maxs = [_np.nan]
            maxs = [_np.nan]

            if winlen < len(signal):
                idx_start = _np.arange(0, len(signal) - winlen + 1, winstep)
            else:
                idx_start = [0]

            for idx_st in idx_start:
                idx_sp = idx_st + winlen
                if idx_sp > len(signal):
                    idx_sp = len(signal)
                curr_win = signal[idx_st: idx_sp]
                curr_idx_max = _np.argmax(curr_win) + idx_st
                curr_max = _np.max(curr_win)

                # peak not already detected & peak not at the beginnig/end of the window:
                if curr_idx_max != idx_maxs[-1] and curr_idx_max != idx_st and curr_idx_max != idx_sp - 1:
                    idx_maxs.append(curr_idx_max)
                    maxs.append(curr_max)
            idx_maxs = idx_maxs[1:]
            maxs = maxs[1:]
            return _np.array(idx_maxs), _np.array(maxs)


class Minima(_Tool):
    """
    Find all local minima in the signal

    Parameters
    ----------
    method : str
        Method to detect the minima. Available methods: 'complete' or 'windowing'. 'complete' finds all the local
        minima, 'windowing' uses a runnning window to find the global minima in each window.
    win_len : float, >0
        Length of window in seconds (method = 'windowing')
    win_step : float, >0
        Shift of the window to start the next window in seconds (method = 'windowing')

    Optional parameters
    -------------------
    refractory : float, >0, default = 0
        Seconds to skip after a detected minimum to look for new minima, when method = 'complete'. 

    Returns
    -------
    idx_mins : array
        Array containing indexes of the minima
    val_mins : array
        Array containing values of the minima
    """

    def __init__(self, method='complete', refractory=0, win_len=None, win_step=None):
        assert method in ['complete', 'windowing'], "Method not valid"
        assert refractory >= 0, "Refractory time value should be positive (or 0 to deactivate)"
        
        if method == 'windowing':
            assert win_len > 0, "Window length should be positive"
            assert win_step > 0, "Window step should be positive"
            _Tool.__init__(self, method=method, refractory=refractory, win_len=win_len, win_step=win_step)
        elif method == 'complete':
            _Tool.__init__(self, method=method, refractory=refractory)

    @classmethod
    def algorithm(cls, signal, params):
        idx_mins, mins = Maxima(**params)(-signal.copy())
        return idx_mins, -1 * mins


class CreateTemplate(_Tool):
    """
    Create a template for matched filtering
    
    Parameters
    ----------
    ref_indexes : array of int
        Indexes of the signals to be used as reference point to generate the template
    idx_start : int, >0
        Index of the signal to start the segmentation of the portion used to generate the template
    idx_stop : int, >0
        Index of the signal to end the segmentation of the portion used to generate the template
    smp_pre : int, >0
        Number of samples before the reference point to be used to generate the template
    smp_post : int, >0
        Number of samples after the reference point to be used to generate the template
    
    Returns
    -------
    template : numpy.array
        The template
    """

    def __init__(self, ref_indexes, smp_pre, smp_post, idx_start=0, idx_stop=None):
        ref_indexes = _np.array(ref_indexes)
        assert ref_indexes.ndim == 1, "Array of indices has to be 1-dimensional"
        assert smp_pre > 0, "Number of samples before the reference should be positive"
        assert smp_post > 0, "Number of samples after the reference should be positive"
        assert idx_start >= 0, "Start index should be non negative"
        assert idx_stop >= 0, "Stop index should be positive or 0 for end of the signal"
        _Tool.__init__(self, ref_indexes=ref_indexes, smp_pre=smp_pre, smp_post=smp_post, idx_start=idx_start,
                       idx_stop=idx_stop)

    @classmethod
    def algorithm(cls, signal, params):
        ref_indexes = params['ref_indexes']
        idx_start = params['idx_start']
        idx_end = params['idx_stop']
        if idx_end is None:
            idx_end = len(signal)
        smp_pre = params['smp_pre']
        smp_post = params['smp_post']

        sig = _np.array(signal[idx_start: idx_end])
        ref_indexes = _np.array(
            ref_indexes[_np.where((ref_indexes > idx_start) & (ref_indexes <= idx_end))[0]]) - idx_start
        total_samples = smp_pre + smp_post
        templates = _np.zeros(total_samples)

        for i in range(1, len(ref_indexes) - 1):
            idx_peak = ref_indexes[i]
            tmp = sig[idx_peak - smp_pre: idx_peak + smp_post]
            tmp = (tmp - _np.min(tmp)) / (_np.max(tmp) - _np.min(tmp))
            templates = _np.c_[templates, tmp]

        templates = templates[:, 1:]

        template = _np.mean(templates, axis=1)
        template = template - template[0]
        template = template / _np.sum(template)
        return template


class BootstrapEstimation(_Tool):
    """
    Perform a bootstrapped estimation of given statistical indicator
    
    Parameters
    ----------
    func : numpy function
        Function to use in the bootstrapping. Must accept data as input
        
    Optional parameters
    -------------------
    
    n : int, >0, default = 100
        Number of iterations
    k : float, (0,1), default = 0.5
        Portion of data to be used at each iteration
    
    Returns
    -------
    estim : float
        Bootstrapped estimate
    
    """

    def __init__(self, func, n=100, k=0.5):
        from types import FunctionType as Func
        assert isinstance(func, Func), "Parameter function should be a function (types.FunctionType)"
        assert n > 0, "n should be positive"
        assert 0 < k <= 1, "k should be between (0 and 1]"
        _Tool.__init__(self, func=func, n=n, k=k)

    @classmethod
    def algorithm(cls, signal, params):
        signal = _np.asarray(signal)
        l = len(signal)
        func = params['func']
        niter = int(params['n'])
        k = params['k']

        estim = []
        for i in range(niter):
            ixs = _np.arange(l)
            ixs_p = _np.random.permutation(ixs)
            sampled_data = signal[ixs_p[:int(round(k * l))]]
            curr_est = func(sampled_data)
            estim.append(curr_est)
        estim = _np.sort(estim)
        return estim[int(len(estim) / 2)]


class Durations(_Tool):
    """
    Compute durations of events starting from their start and stop indexes

    Parameters:
    -----------
    starts : array
        Start indexes along the data
    stops : array
        Stop indexes along the data

    Return:
    -------
    durations : array
        durations of the events
    """

    def __init__(self, starts, stops):
        starts = _np.array(starts)
        assert starts.ndim == 1
        stops = _np.array(stops)
        assert stops.ndim == 1
        _Tool.__init__(self, starts=starts, stops=stops)

    @classmethod
    def algorithm(cls, signal, params):
        starts = params["starts"]
        stops = params["stops"]

        fsamp = signal.get_sampling_freq()
        durations = []
        for I in range(len(starts)):
            if (stops[I] > 0) & (starts[I] >= 0):
                durations.append((stops[I] - starts[I]) / fsamp)
            else:
                durations.append(_np.nan)
        return durations


class Slopes(_Tool):
    """
    Compute rising slope of peaks

    Parameters:
    -----------
    starts : array
        Start of the peaks indexes
    peaks : array
        Peaks indexes

    Return:
    -------
    slopes : array
        Rising slopes the peaks
    """

    def __init__(self, starts, peaks):
        starts = _np.array(starts)
        assert starts.ndim == 1
        peaks = _np.array(peaks)
        assert peaks.ndim == 1
        _Tool.__init__(self, starts=starts, peaks=peaks)

    @classmethod
    def algorithm(cls, data, params):
        starts = params["starts"]
        peaks = params["peaks"]

        fsamp = data.get_sampling_freq()
        slopes = []
        for I in range(len(starts)):
            if peaks[I] > 0 & starts[I] >= 0:
                dy = data[peaks[I]] - data[starts[I]]
                dt = (peaks[I] - starts[I]) / fsamp
                slopes.append(dy / dt)
            else:
                slopes.append(_np.nan)
        return slopes


class BeatOutliers(_Tool):
    """
    Detects outliers in the IBI signal. 
    
    Optional parameters
    -------------------
    
    cache : int, >0,  default=3
        Number of IBI to be stored in the cache for adaptive computation of the interval of accepted values
    sensitivity : float, >0, default = 0.25
        Relative variation from the current IBI median value of the cache that is accepted
    ibi_median : float, >=0, default = 0
        IBI value use to initialize the cache. By default (ibi_median=0) it is computed as median of the input IBI
    
    Returns
    -------
    id_bad_ibi : numpy.array
        Identifiers of wrong beats
    
    Notes
    -----
    It only detects outliers. You should manually remove outliers using FixIBI
    
    """

    def __init__(self, ibi_median=0, cache=3, sensitivity=0.25):
        assert ibi_median >= 0, "IBI median value should be positive (or equal to 0 for automatic computation"
        assert cache >= 1, "Cache size should be greater than 1"
        assert sensitivity > 0, "Sensitivity value shlud be positive"

        _Tool.__init__(self, ibi_median=ibi_median, cache=cache, sensitivity=sensitivity)

    @classmethod
    def get_signal_type(cls):
        return ['IBI']

    @classmethod
    def algorithm(cls, signal, params):
        cache, sensitivity, ibi_median = params["cache"], params["sensitivity"], params["ibi_median"]

        if ibi_median == 0:
            ibi_expected = float(_np.median(signal))
        else:
            ibi_expected = float(ibi_median)

        id_bad_ibi = []
        ibi_cache = _np.repeat(ibi_expected, cache)
        counter_bad = 0

        # missings = []
        idx_ibi = signal.get_indices()
        ibi = signal.get_values()
        for i in range(1, len(idx_ibi)):
            curr_median = _np.median(ibi_cache)

            curr_ibi = ibi[i]

            if curr_ibi > curr_median * (1 + sensitivity):  # abnormal peak:
                id_bad_ibi.append(i)  # append ibi id to the list of bad ibi
                counter_bad += 1
            # missings.append([idx_ibi[i-1],idx_ibi[i]])

            elif curr_ibi < curr_median * (1 - sensitivity):  # abnormal peak:
                id_bad_ibi.append(i)  # append ibi id to the list of bad ibi
                counter_bad += 1
            else:
                ibi_cache = _np.r_[ibi_cache[1:], curr_ibi]
                counter_bad = 0
            if counter_bad == cache:  # ibi cache probably corrupted, reinitialize
                ibi_cache = _np.repeat(ibi_expected, cache)
                counter_bad = 0

        return id_bad_ibi


class FixIBI(_Tool):
    """
    Corrects the IBI series removing abnormal IBI
    
    Parameters
    ----------
    idx_bad_ibi : array
        Identifiers of abnormal beats
   
    Returns
    -------
    ibi : Unevenly Signal
        Corrected IBI
            
    """

    def __init__(self, idx_bad_ibi):
        idx_bad_ibi = _np.array(idx_bad_ibi)
        assert idx_bad_ibi.ndim == 1
        _Tool.__init__(self, id_bad_ibi=idx_bad_ibi)

    @classmethod
    def get_signal_type(cls):
        return ['IBI']

    @classmethod
    def algorithm(cls, signal, params):
        assert isinstance(signal,
                          _UnevenlySignal), "IBI can only be represented by an UnevenlySignal, %s found." % type(signal)
        id_bad = params['id_bad_ibi']
        idx_ibi = signal.get_indices()
        ibi = signal.get_values()
        idx_ibi_nobad = _np.delete(idx_ibi, id_bad)
        ibi_nobad = _np.delete(ibi, id_bad)
        idx_ibi = idx_ibi_nobad.astype(int)
        ibi = ibi_nobad
        return _UnevenlySignal(values = ibi, 
                               sampling_freq = signal.get_sampling_freq(), 
                               start_time = signal.get_start_time(),
                               signal_type = signal.get_signal_type(), 
                               x_values=idx_ibi, x_type='indices', 
                               duration=signal.get_duration())


class BeatOptimizer(_Tool):
    """
    Optimize detection of errors in IBI estimation.
    
    Optional parameters
    -------------------

    B : float, >0, default = 0.25
        Ball radius in seconds to allow pairing between forward and backward beats
    cache : int, >0,  default = 3
        Number of IBI to be stored in the cache for adaptive computation of the interval of accepted values
    sensitivity : float, >0, default = 0.25
        Relative variation from the current IBI median value of the cache that is accepted
    ibi_median : float, >=0, default = 0
        IBI value use to initialize the cache. By default (ibi_median=0) it is computed as median of the input IBI

    Returns
    -------
    ibi : UnevenlySignal
        Optimized IBI signal

    Notes
    -----
        Bizzego et al., *DBD-RCO: Derivative Based Detection and Reverse Combinatorial Optimization 
        to improve heart beat detection for wearable devices for info about the algorithm*
    """

    def __init__(self, b=0.25, ibi_median=0, cache=3, sensitivity=0.25):
        assert b > 0, "Ball radius should be positive"
        assert ibi_median >= 0, "IBI median value should be positive (or equal to 0 for automatic computation"
        assert cache >= 1, "Cache size should be greater than 1"
        assert sensitivity > 0, "Sensitivity value shlud be positive"

        _Tool.__init__(self, B=b, ibi_median=ibi_median, cache=cache, sensitivity=sensitivity)

    @classmethod
    def get_signal_type(cls):
        return ['IBI']

    @classmethod
    def algorithm(cls, signal, params):
        cls.error('Not implemented')
        return signal

        b, cache, sensitivity, ibi_median = params["B"], params["cache"], params["sensitivity"], params["ibi_median"]

        fsamp = signal.get_sampling_freq()

        # FORWARD
        id_bad_ibi_F = BeatOutliers(ibi_median=ibi_median, cache=cache, sensitivity=sensitivity)(signal)

        # BACKWARD
        # reverse the signal
        idx_reverse = signal.get_indices()

        idx_reverse = -1 * (idx_reverse - idx_reverse[-1])[::-1]
        ibi_reverse = _np.diff(idx_reverse)
        ibi_reverse = _np.r_[ibi_reverse[0], ibi_reverse]

        ibi_reverse = _UnevenlySignal(ibi_reverse / fsamp, x_values=idx_reverse, x_type='indices',
                                      duration=signal.get_duration())

        id_bad_ibi_b = BeatOutliers(ibi_median=ibi_median, cache=cache, sensitivity=sensitivity)(ibi_reverse)

        id_bad_ibi_B = -1 * (_np.array(id_bad_ibi_b) - len(ibi_reverse))[::-1]

        # idx_1
        ###
        # add indexes of idx_ibi_2 which are not in idx_ibi_1 but close enough
        b = b * fsamp
        for i_2 in range(1, len(idx_2)):
            curr_idx_2 = idx_2[i_2]
            if not (curr_idx_2 in idx_1):
                i_1 = _np.where((idx_1 >= curr_idx_2 - b) & (idx_1 <= curr_idx_2 + b))[0]
                if not len(i_1) > 0:
                    idx_1 = _np.r_[idx_1, curr_idx_2]
        idx_1 = _np.sort(idx_1)

        ###
        # create pairs for each beat
        pairs = [[0, 0]]
        for i_1 in range(1, len(idx_1)):
            curr_idx_1 = idx_1[i_1]
            if curr_idx_1 in idx_2:
                pairs.append([curr_idx_1, curr_idx_1])
            else:
                i_2 = _np.where((idx_2 >= curr_idx_1 - b) & (idx_2 <= curr_idx_1 + b))[0]
                if len(i_2) > 0:
                    i_2 = i_2[0]
                    pairs.append([curr_idx_1, idx_2[i_2]])
                else:
                    pairs.append([curr_idx_1, curr_idx_1])
        pairs = _np.array(pairs)

        ########################################
        # define zones where there are different values
        diff_idxs = pairs[:, 0] - pairs[:, 1]
        diff_idxs[diff_idxs != 0] = 1
        diff_idxs = _np.diff(diff_idxs)

        starts = _np.where(diff_idxs > 0)[0]
        stops = _np.where(diff_idxs < 0)[0]

        if len(starts) == 0:  # no differences
            # keep the 'outliers removed' version
            idx_1 = idx_1 + idx_st
            return _UnevenlySignal(ibi_1 / fsamp, sampling_freq=fsamp, signal_type="IBI",
                                   start_time=signal.get_start_time(), x_values=idx_1, x_type='indices',
                                   duration=signal.get_duration())

        if len(stops) == 0:
            stops = _np.array([starts[-1] + 1])

        if starts[0] >= stops[0]:
            stops = stops[1:]

        stops += 1

        if len(starts) > len(stops):
            stops = _np.r_[stops, starts[-1] + 1]

        # split long sequences
        new_starts = _np.copy(starts)
        new_stops = _np.copy(stops)

        add_index = 0
        lens = stops - starts
        for i in range(len(starts)):
            l = lens[i]
            if l > 10:
                curr_st = starts[i]
                curr_sp = stops[i]
                new_st = _np.arange(curr_st, curr_sp, 4)
                new_sp = new_st + 4
                new_sp[-1] = curr_sp
                new_starts = _np.delete(new_starts, i + add_index)
                new_stops = _np.delete(new_stops, i + add_index)
                new_starts = _np.insert(new_starts, i + add_index, new_st)
                new_stops = _np.insert(new_stops, i + add_index, new_sp)
                add_index = add_index + len(new_st) - 1

        starts = new_starts
        stops = new_stops

        ########################################
        # find best combination
        idx_out = _np.copy(pairs[:, 0])
        for i in range(len(starts)):
            i_st = starts[i]
            i_sp = stops[i]

            if i_sp > len(idx_out) - 1:
                i_sp = len(idx_out) - 1

            curr_portion = _np.copy(pairs[i_st - 1: i_sp + 1, :])

            best_portion = None
            best_error = _np.Inf

            combinations = list(_itertools.product([0, 1], repeat=i_sp - i_st - 1))
            for comb in combinations:
                cand_portion = _np.copy(curr_portion[:, 0])
                for k in range(len(comb)):
                    bit = comb[k]
                    cand_portion[k + 2] = curr_portion[k + 2, bit]
                cand_error = sum(abs(_np.diff(_np.diff(cand_portion))))
                if cand_error < best_error:
                    best_portion = cand_portion
                    best_error = cand_error
            idx_out[i_st - 1: i_sp + 1] = best_portion

        ###
        # finalize arrays
        idx_out = _np.array(idx_out) + idx_st
        ibi_out = _np.diff(idx_out)
        ibi_out = _np.r_[signal.get_values()[0], ibi_out / fsamp]

        return _UnevenlySignal(ibi_out, sampling_freq=signal.get_sampling_freq(), signal_type="IBI",
                               start_time=signal.get_start_time(), x_values=idx_out, x_type='indices',
                               duration=signal.get_duration())


# EDA Tools
class OptimizeBateman(_Tool):
    """
    Optimize the Bateman parameters T1 and T2.
    
    Parameters
    ----------
    delta : float
        Minimum amplitude of the peaks in the driver
        
    Optional parameters
    -------------------
    
    opt_method : str
        Method to perform the search of optimal parameters.
        Available methods:
    * 'bsh' Adaptive Simulated Annealing. Uses the Basin-Hopping algorithm.
    * 'grid' Grid search
    complete : boolean, default = True
        Whether to perform minimization after detecting the optimal parameters
    par_ranges : list, default = [0.1, 0.99, 1, 10]
        [min_T1, max_T1, min_T2, max_T2] boundaries for the Bateman parameters
    maxiter : int (Default = 99999)
        Maximum number of iterations ('asa' method)
    n_step_1 : int
        Number of steps in the grid search (paramter t1)
    n_step_2 : int
        Number of steps in the grid search (parameter t2)
    weight : str
        How the errors should be weighted before computing the loss function. ['exp', 'lin', 'none']
    min_pars : dict
        Additional parameters to pass to the minimization function (when complete = True)

    Returns
    -------
    x0 : list
        The resulting optimal parameters
    
    """

    # TODO (Feature): add **kwargs parameters for internal minimization
    def __init__(self, delta, loss_func='all', opt_method='bsh', complete=False, par_ranges=None,
                 maxiter=99999, n_step_1=10, n_step_2=10, **kwargs):
        if par_ranges is None:
            par_ranges = [0.1, 0.99, 1.5, 10]
        assert delta > 0
        assert loss_func in ['ben', 'all'], "Loss function not valid"
        assert opt_method in ['grid', 'bsh'], "Optimization method not valid"
        assert len(par_ranges) == 4
        if opt_method == "bsh":
            assert maxiter > 0
        if opt_method == "grid":
            assert n_step_1 > 0
            assert n_step_2 > 0

        _Tool.__init__(self, delta=delta, loss_func=loss_func, opt_method=opt_method, complete=complete,
                       par_ranges=par_ranges, maxiter=maxiter, n_step_1=n_step_1, n_step_2=n_step_2,
                       **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        delta = params['delta']
        opt_method = params['opt_method']
        complete = params['complete']
        par_ranges = params['par_ranges']
        maxiter = params['maxiter']
        n_step_1 = params['n_step_1']
        n_step_2 = params['n_step_2']

        # TODO (Andrea):explain "add **kwargs"

        if params['loss_func'] == 'ben':
            loss_function = OptimizeBateman._loss_benedek
        elif params['loss_func'] == 'all':
            loss_function = OptimizeBateman._loss_function_all

        if opt_method == 'grid':
            min_T1 = float(par_ranges[0])
            max_T1 = float(par_ranges[1])
    
            min_T2 = float(par_ranges[2])
            max_T2 = float(par_ranges[3])
            step_T1 = (max_T1 - min_T1) / n_step_1
            step_T2 = (max_T2 - min_T2) / n_step_2
            rranges = (slice(min_T1, max_T1 + step_T1, step_T1), slice(min_T2, max_T2 + step_T2, step_T2))
            x0, loss, grid, loss_grid = _opt.brute(loss_function, rranges,
                                                   args=(signal, delta),
                                                   full_output=True, finish=None)
            exit_code = -1

        elif opt_method == 'bsh':
            x_opt = _opt.basinhopping(loss_function, [0.75, 2.],
                                      niter=maxiter,
                                      minimizer_kwargs={
                                          "bounds": ((par_ranges[0], par_ranges[1]), (par_ranges[2], par_ranges[3])),
                                          "args": (signal, delta)},
                                      disp=False, niter_success=10)
            x0 = x_opt.x
            loss = float(x_opt.fun)

            if x_opt.minimization_failures == maxiter:
                exit_code = 1
            else:
                exit_code = 0
        else:
            cls.error("opt_method not understood")
            return None

        if complete:
            x0_min, loss_min, niter,\
            nfuncalls, warnflag, allvec = _opt.fmin(loss_function, x0,
                                                    args=(signal, delta),
                                                    full_output=True)
            return x0, x0_min, loss, loss_min, exit_code, warnflag
        else:
            return x0, loss, exit_code


    @staticmethod
    def _loss_function_all(par_bat, signal, delta):
        """
        Computes the loss for optimization of Bateman parameters.

        Parameters
        ----------
        par_bat : list
            Bateman parameters to be optimized
        signal : Signal
            The EDA signal
        delta : float
            Minimum amplitude of the peaks in the driver
       
        Returns
        -------
        loss : float
            The computed loss
        """

        from ..estimators.Estimators import DriverEstim as _DriverEstim
        # check if pars hit boudaries
        # if par_bat[0] < min_T1 or par_bat[0] > max_T1 or par_bat[1] < min_T2 or par_bat[1] > max_T2 or par_bat[0] >=
        #  par_bat[1]:
        #            return 10000

        REC_TIME = 20  #max recovery time
        if _np.isnan(par_bat[0]) | _np.isnan(par_bat[1]):
            return _np.Inf

        fsamp = signal.get_sampling_freq()

        driver = _DriverEstim(t1=par_bat[0], t2=par_bat[1])(signal)
        maxp, minp, ignored, ignored = PeakDetection(delta=delta, start_max=True)(driver.segment_time(driver.get_start_time(), driver.get_end_time() - REC_TIME ))
        
        if len(maxp) == 0:
            OptimizeBateman.warn('Unable to find peaks in driver signal for computation of Energy. Returning Inf')
            return _np.inf
        else:
            energy = 0

            for idx_max in maxp:
                # extract WLEN seconds after the peak
                driver_portion = driver.segment_idx(idx_max, idx_max + REC_TIME * fsamp)
                
                # find angular coefficient (average derivative)
                

                diff_y = Diff()(driver_portion)
                th_66 = _np.percentile(diff_y, 66)
                th_33 = _np.percentile(diff_y, 33)

                idx_sel_diff_y = _np.where((diff_y > th_33) & (diff_y < th_66))[0]
                diff_y_sel = diff_y.get_values()[idx_sel_diff_y]
                mean_s = BootstrapEstimation(func=_np.mean, n=10, k=0.5)(_UnevenlySignal(diff_y_sel, sampling_freq = fsamp, x_values = idx_sel_diff_y, x_type = 'indices'))
                
                # find intercept
                mean_y = BootstrapEstimation(func=_np.median, n=10, k=0.5)(driver_portion)
                b_mean_s = mean_y - mean_s * len(driver_portion) / 2

                # compute linear trend
                line_mean_s = mean_s * _np.arange(len(driver_portion)) + b_mean_s

                # detrend
                driver_detrended = driver_portion - line_mean_s

                # normalize detrended driver
                driver_detrended /= float(_np.max(driver_detrended))

                energy_curr = (1 / fsamp) * _np.sum(driver_detrended[1:] ** 2) / (len(driver_detrended) - 1)
                energy += energy_curr
            
            energy /= len(maxp)
            OptimizeBateman.log(
                '\r\r ALL. Current parameters: ' + str(par_bat[0]) + ' - ' + str(par_bat[1]) + ' Loss: ' + str(
                    energy) + '\r')
            if _np.isnan(energy):
                return _np.Inf

            return energy

    @staticmethod
    def _loss_benedek(par_bat, signal, delta):
        """
        Computes the loss for optimization of Bateman parameters according to Benedek2010 (REF).
        #TODO: insert reference

        Parameters
        ----------
        par_bat : list
            Bateman parameters to be optimized
        signal : _Signal
            The EDA signal
        delta : float
            Minimum amplitude of the peaks in the driver
       
        Returns
        -------
        loss : float
            The computed loss
        """

        from ..estimators.Estimators import DriverEstim as _DriverEstim

        if _np.isnan(par_bat[0]) | _np.isnan(par_bat[1]):
            return _np.Inf

        def phasic_estim_benedek(driver, delta):
            # find peaks in the driver
            fsamp = driver.get_sampling_freq()
            i_peaks, idx_min, val_max, val_min = PeakDetection(delta=delta, refractory=1, start_max=True)(driver)

            i_pre_max = 10 * fsamp
            i_post_max = 10 * fsamp

            i_start = _np.empty(len(i_peaks), int)
            i_stop = _np.empty(len(i_peaks), int)

            if len(i_peaks) == 0:
                OptimizeBateman.warn('No peaks found.')
                return driver

            for i in range(len(i_peaks)):
                i_pk = int(i_peaks[i])

                # find START
                i_st = i_pk - i_pre_max
                if i_st < 0:
                    i_st = 0

                driver_pre = driver[i_st:i_pk]
                i_pre = len(driver_pre) - 2

                while i_pre > 0 and (driver_pre[i_pre] >= driver_pre[i_pre - 1]):
                    i_pre -= 1

                i_start[i] = i_st + i_pre + 1

                # find STOP
                i_sp = i_pk + i_post_max
                if i_sp >= len(driver):
                    i_sp = len(driver) - 1

                driver_post = driver[i_pk: i_sp]
                i_post = 1

                while i_post < len(driver_post) - 2 and (driver_post[i_post] >= driver_post[i_post + 1]):
                    i_post += 1

                i_stop[i] = i_pk + i_post

            idxs_peak = _np.array([])

            for i_st, i_sp in zip(i_start, i_stop):
                idxs_peak = _np.r_[idxs_peak, _np.arange(i_st, i_sp)]

            idxs_peak = idxs_peak.astype(int)

            # generate the grid for the interpolation
            idx_grid_candidate = _np.arange(0, len(driver) - 1, 10 * fsamp)

            idx_grid = []
            for i in idx_grid_candidate:
                if i not in idxs_peak:
                    idx_grid.append(i)

            if len(idx_grid) == 0:
                idx_grid.append(0)

            if idx_grid[0] != 0:
                idx_grid = _np.r_[0, idx_grid]
            if idx_grid[-1] != len(driver) - 1:
                idx_grid = _np.r_[idx_grid, len(driver) - 1]

            driver_grid = _UnevenlySignal(driver[idx_grid], fsamp, "dEDA", driver.get_start_time(), x_values=idx_grid,
                                          x_type='indices', duration=signal.get_duration())
            if len(idx_grid) >= 4:
                tonic = driver_grid.to_evenly(kind='cubic')
            else:
                tonic = driver_grid.to_evenly(kind='linear')
            phasic = driver - tonic
            return phasic

        ALPHA = 6
        driver = _DriverEstim(t1=par_bat[0], t2=par_bat[1])(signal)

        phasic = phasic_estim_benedek(driver, delta)

        TH = float(_np.max(phasic) * 0.05)

        peaks = _np.zeros(len(phasic))
        peaks[phasic > TH] = 1

        # compute indistinctness
        d_peaks_dt = _np.diff(peaks)

        starts = _np.where(d_peaks_dt == 1)[0]
        ends = _np.where(d_peaks_dt == -1)[0]

        if (len(starts) == 0) | (len(ends) == 0):
            indist = _np.Inf
        else:
            if ends[0] < starts[0]:  # phasic starts with a peak
                starts = _np.r_[0, starts]

            if ends[-1] < starts[-1]:  # phasic ends with a peak
                ends = _np.r_[ends, len(peaks) - 1]

            # at this point is should be len(starts)==len(ends)
            indist = _np.sum((ends - starts) ** 2) / phasic.get_duration()

        # compute negativeness
        negatives = phasic[phasic < 0]
        neg = _np.sqrt(_np.mean(negatives ** 2))

        # compute loss
        LOSS = indist + ALPHA * neg
        OptimizeBateman.log(
            '\r\r BENEDEK. Current parameters: ' + str(par_bat[0]) + ' - ' + str(par_bat[1]) + ' Loss: ' + str(
                LOSS) + '\r')
        return LOSS



# coding=utf-8
from __future__ import division

from . import ph, np, TestData, approx


# %%
def test_tools():
    # %%
    FSAMP = 2048
    TSTART = 0

    # %%
    bvp = ph.EvenlySignal(TestData.bvp(), sampling_freq=FSAMP, signal_nature='BVP', start_time=TSTART)

    # %%
    # TEST SignalRange
    rng_bvp = ph.SignalRange(win_len=1, win_step=0.5, smooth=True)(bvp)  # OK
    assert (int(np.max(rng_bvp) * 100) == 1746)

    # %%
    # TEST PeakDetection
    idx_mx, idx_mn, mx, mn = ph.PeakDetection(delta=rng_bvp * 0.5)(bvp)

    assert (np.sum(idx_mx) == 16818171)
    assert (np.sum(idx_mn) == 16726017)

    # %%
    # EDA
    eda = ph.EvenlySignal(TestData.eda(), sampling_freq=FSAMP, signal_nature='EDA', start_time=TSTART)
    eda = eda.resample(fout=8)

    eda = eda.resample(8)

    # filter
    eda = ph.IIRFilter(fp=0.8, fs=1.1)(eda)

    driver = ph.DriverEstim()(eda)
    phasic, _, __ = ph.PhasicEstim(delta=0.02)(driver)

    idx_mx, idx_mn, mx, mn = ph.PeakDetection(delta=0.02)(phasic)

    # TEST PeakSelection
    st, sp = ph.PeakSelection(indices=idx_mx, win_pre=2, win_post=2)(phasic)
    assert (np.sum(st) == 22985)

    # %%
    # TEST PSD
    FSAMP = 100
    n = np.arange(1000)
    t = n / FSAMP
    freq = 2.5

    sinusoid = ph.EvenlySignal(np.sin(2 * np.pi * freq * t), sampling_freq=FSAMP, signal_nature='', start_time=0)

    f, psd = ph.PSD(method='welch', nfft=4096, window='hanning')(sinusoid)

    assert (f[np.argmax(psd)] == approx(2.5, abs=0.02))

    sinusoid_unevenly = ph.UnevenlySignal(np.delete(sinusoid.get_values(), np.arange(10, 200)),
                                          sampling_freq=FSAMP,
                                          start_time=0,
                                          x_values=np.delete(sinusoid.get_times(), np.arange(10, 200)),
                                          x_type='instants')

    f, psd = ph.PSD(method='welch', nfft=4096, window='hanning')(sinusoid_unevenly)

    assert (f[np.argmax(psd)] == approx(2.5, abs=0.02))

    # %%
    # TEST Maxima
    idx_mx, mx = ph.Maxima(method='windowing', win_len=1, win_step=0.5)(bvp)
    assert (np.sum(idx_mx) == 16923339)

    # TEST Minima
    idx_mn, mn = ph.Minima(method='windowing', win_len=1, win_step=0.5)(bvp)
    assert (np.sum(idx_mn) == 17939276)



# coding=utf-8
import os as _os
try:
    import pyphysio as ph
except ImportError:
    from sys import path
    path.insert(0, _os.path.join(_os.path.dirname(__file__), ".."))
    import pyphysio as ph
import numpy as np
# noinspection PyUnresolvedReferences

#from pytest import approx as approx

__author__ = 'aleb'


class TestData(object):
    _sing = None
    _path = _os.path.join(_os.path.dirname(__file__), "data")
    _file = "medical.txt.bz2"

    @classmethod
    def get_data(cls):
        if TestData._sing is None:
            TestData._sing = np.genfromtxt(_os.path.join(TestData._path, TestData._file), delimiter="\t")
        return TestData._sing

    # The following methods return an array to make it easier to test the Signal wrapping classes

    @classmethod
    def ecg(cls):
        return TestData.get_data()[:, 0]

    @classmethod
    def eda(cls):
        return TestData.get_data()[:, 1]

    @classmethod
    def bvp(cls):
        return TestData.get_data()[:, 2]

    @classmethod
    def resp(cls):
        return TestData.get_data()[:, 3]

#TODO: Add timeline


# coding=utf-8
from __future__ import division

from ..BaseIndicator import Indicator as _Indicator
from ..tools.Tools import Diff as _Diff
from ..indicators.TimeDomain import Mean as _Mean, StDev as _StDev
from scipy.spatial.distance import cdist as _cd
import numpy as _np

__author__ = 'AleB'


class PoincareSD1(_Indicator):
    """
    Return the SD1 value of the Poincare' plot of input Inter Beat Intervals

    Returns
    -------
    sd1 : float
        SD1 of Poincare' plot
    
    """

    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        """
        Calculates Poincare SD 1 and 2
        @return: (SD1, SD2)
        @rtype: (array, array)
        """
        xd, yd = _np.array(list(data[:-1])), _np.array(list(data[1:]))
        sd1 = _np.std((xd - yd) / _np.sqrt(2.0))
        return sd1


class PoincareSD2(_Indicator):
    """
    Return the SD2 value of the Poincare' plot of input Inter Beat Intervals

    Returns
    -------
    sd2 : float
        SD2 of Poincare' plot
    
    """

    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        """
        Calculates Poincare SD 1 and 2
        @return: (SD1, SD2)
        @rtype: (array, array)
        """
        xd, yd = _np.array(list(data[:-1])), _np.array(list(data[1:]))
        sd2 = _np.std((xd + yd) / _np.sqrt(2.0))
        return sd2


class PoincareSD1SD2(_Indicator):
    """
    Return the SD1/SD2 value of the Poincare' plot of input Inter Beat Intervals

    Returns
    -------
    sd12 : float
        SD1/SD2 of Poincare' plot
    
    """

    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        """
        Calculates Poincare SD 1 and 2
        @return: (SD1, SD2)
        @rtype: (array, array)
        """
        sd1 = PoincareSD1()(data)
        sd2 = PoincareSD2()(data)
        return sd1 / sd2


class PoinEll(_Indicator):
    """
    Return the SD1*SD2*pi value of the Poincare' plot of input Inter Beat Intervals

    Returns
    -------
    sdEll : float
        SD1*SD2*pi of Poincare' plot
    
    """

    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        sd1 = PoincareSD1()(data)
        sd2 = PoincareSD2()(data)
        return sd1 * sd2 * _np.pi


class PNNx(_Indicator):
    """
    Computes the relative frequency of pairs of consecutive samples s1, s2 such that s1-s2 >= 'threshold' in
    milliseconds.
    
    Parameters
    ----------
    threshold : int, >0
        Threshold in milliseconds.
        
    Returns
    -------
    PNNx : float
        Relative frequency
    """

    def __init__(self, threshold, **kwargs):
        _Indicator.__init__(self, threshold=threshold, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        return NNx.algorithm(data, params) / float(len(data))


class NNx(_Indicator):
    """
    Counts the pairs of consecutive samples s1, s2 such that s1-s2 >= 'threshold' in milliseconds.
    
    Parameters
    ----------
    threshold : int, >0
        Threshold in milliseconds.
    """

    def __init__(self, threshold, **kwargs):
        assert threshold > 0, "Not implemented for threshold not > 0"
        _Indicator.__init__(self, threshold=threshold, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        th = params['threshold']
        diff = _Diff()(signal)
        return sum(1.0 for x in diff * 1000 if x > th)


class _Embed(_Indicator):
    def __init__(self, dimension, **kwargs):
        _Indicator.__init__(self, dimension=dimension, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        """
        Calculates the the vector of the sequences of length 'subset_size' of the data
        @return: Data array with shape (l - n + 1, n) having l=len(data) and n=subset_size
        @rtype: array
        """
        n = params['dimension']
        # t = params['delay']
        num = len(signal) - n + 1
        if num > 0:
            emb = _np.zeros([num, n])
            # 2>3 xrange>range
            for i in range(num):
                emb[i, :] = signal[i:i + n]
            return emb
        else:
            return []


class ApproxEntropy(_Indicator):
    """
    Calculates Approximate Entropy
        
    Optional Parameters
    ----------
    radius : float, >0, default=0.5
        Radius to threshold the distance between the embedded vectors
        
    Returns
    -------
    apen : float
        Approximate Entropy
    """

    def __init__(self, radius=.5, **kwargs):
        assert radius > 0, "Parameter radius should be > 0"
        _Indicator.__init__(self, radius=radius, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        if len(data) < 3:
            return _np.nan
        else:
            r = params['radius']
            uj_m = _Embed(dimension=2, delay=1)(data)
            uj_m1 = _Embed(dimension=3, delay=1)(data)
            card_elem_m = uj_m.shape[0]
            card_elem_m1 = uj_m1.shape[0]

            r = r * _np.std(data)
            d_m = _cd(uj_m, uj_m, 'chebyshev')
            d_m1 = _cd(uj_m1, uj_m1, 'chebyshev')

            cmr_m_ap_en = _np.zeros(card_elem_m)
            # 2>3 xrange>range
            for i in range(card_elem_m):
                vector = d_m[i]
                cmr_m_ap_en[i] = float(sum(1 for i in vector if i <= r)) / card_elem_m

            cmr_m1_ap_en = _np.zeros(card_elem_m1)
            # 2>3 xrange>range
            for i in range(card_elem_m1):
                vector = d_m1[i]
                cmr_m1_ap_en[i] = float(sum(1 for i in vector if i <= r)) / card_elem_m1

            phi_m = _np.sum(_np.log(cmr_m_ap_en)) / card_elem_m
            phi_m1 = _np.sum(_np.log(cmr_m1_ap_en)) / card_elem_m1

            return phi_m - phi_m1


class SampleEntropy(_Indicator):
    """
    Calculates Sample Entropy
        
    Optional Parameters
    ----------
    radius : float, >0, default=0.5
        Radius to threshold the distance between the embedded vectors
        
    Returns
    -------
    sampen : float
        Sample Entropy
    """

    def __init__(self, radius=.5, **kwargs):
        assert radius > 0, "Parameter radius should be > 0"
        _Indicator.__init__(self, radius=radius, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        if len(data) < 4:
            return _np.nan
        else:
            r = params['radius']
            uj_m = _Embed(dimension=2, delay=1)(data)
            uj_m1 = _Embed(dimension=3, delay=1)(data)

            num_elem_m = uj_m.shape[0]
            num_elem_m1 = uj_m1.shape[0]

            r = r * _StDev()(data)

            d_m = _cd(uj_m, uj_m, 'chebyshev')
            d_m1 = _cd(uj_m1, uj_m1, 'chebyshev')

            cmr_m_sa_mp_en = _np.zeros(num_elem_m)
            # 2>3 xrange>range
            for i in range(num_elem_m):
                vector = d_m[i]
                cmr_m_sa_mp_en[i] = (sum(1 for i in vector if i <= r) - 1) / (num_elem_m - 1)

            cmr_m1_sa_mp_en = _np.zeros(num_elem_m1)
            # 2>3 xrange>range
            for i in range(num_elem_m1):
                vector = d_m1[i]
                cmr_m1_sa_mp_en[i] = (sum(1 for i in vector if i <= r) - 1) / (num_elem_m1 - 1)

            cm = _np.sum(cmr_m_sa_mp_en) / num_elem_m
            cm1 = _np.sum(cmr_m1_sa_mp_en) / num_elem_m1

            return _np.log(cm / cm1)


class DFAShortTerm(_Indicator):
    """
    Calculate the alpha1 (short term) component index of the De-trended Fluctuation Analysis.
   
    Returns
    -------
    alpha1 : float
        Short term component index of the De-trended Fluctuation Analysis
    """

    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):

        x = data
        if len(x) < 16:
            return _np.nan
        else:
            ave = float(_Mean()(x))
            y = _np.cumsum(x).astype(float)
            y -= ave
            l = _np.arange(4, 17, 4)
            f = _np.zeros(len(l))  # f(n) of different given box length n
            # 2>3 xrange>range
            for i in range(0, len(l)):
                n = int(l[i])  # for each box length l[i]
                # 2>3 xrange>range
                for j in range(0, len(x), n):  # for each box
                    if j + n < len(x):
                        c = range(j, j + n)
                        c = _np.vstack([c, _np.ones(n)]).T  # coordinates of time in the box
                        z = y[j:j + n]  # the value of example_data in the box
                        f[i] += _np.linalg.lstsq(c, z)[1]  # add residue in this box
                f[i] /= ((len(x) / n) * n)
            f = _np.sqrt(f)
            return _np.linalg.lstsq(_np.vstack([_np.log(l), _np.ones(len(l))]).T, _np.log(f))[0][0]


class DFALongTerm(_Indicator):
    """
    Calculate the alpha2 (long term) component index of the De-trended Fluctuation Analysis.
   
    Returns
    -------
    alpha2 : float
        Long term component index of the De-trended Fluctuation Analysis
    """

    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        x = data
        if len(x) < 64:
            return _np.nan
        else:
            ave = float(_Mean()(x))
            y = _np.cumsum(x).astype(float)
            y -= ave
            l_max = _np.min([64, len(x)])
            l = _np.arange(16, l_max + 1, 4)
            f = _np.zeros(len(l))  # f(n) of different given box length n
            # 2>3 xrange>range
            for i in range(0, len(l)):
                n = int(l[i])  # for each box length l[i]
                # 2>3 xrange>range
                for j in range(0, len(x), n):  # for each box
                    if j + n < len(x):
                        c = range(j, j + n)
                        c = _np.vstack([c, _np.ones(n)]).T  # coordinates of time in the box
                        z = y[j:j + n]  # the value of example_data in the box
                        f[i] += _np.linalg.lstsq(c, z)[1]  # add residue in this box
                f[i] /= ((len(x) / n) * n)
            f = _np.sqrt(f)

            return _np.linalg.lstsq(_np.vstack([_np.log(l), _np.ones(len(l))]).T, _np.log(f))[0][0]



# coding=utf-8
from __future__ import division

from abc import abstractmethod as _abstract, ABCMeta as _ABCMeta

import numpy as _np
from ..BaseIndicator import Indicator as _Indicator
from ..tools.Tools import PeakDetection as _PeakDetection, PeakSelection as _PeakSelection, Durations as _Durations, \
    Slopes as _Slopes

__author__ = 'AleB'


class _Peaks(_Indicator):
    """
    Peaks base class
    """
    __metaclass__ = _ABCMeta

    def __init__(self, delta, **kwargs):
        assert delta > 0, 'Parameter delta, i.e. amplitude of the minimum peak, has to be > 0'
        _Indicator.__init__(self, delta=delta, **kwargs)

    @classmethod
    @_abstract
    def algorithm(cls, data, params):
        pass


class PeaksMax(_Peaks):
    """
    Return the maximum amplitude of detected peaks.

    Parameters
    ----------
    delta : float, >0
        Minimum amplitude of peaks to be selected
    
    Returns
    -------
    mx : float
        Maximum amplitude of detected peaks
    
    """
    def __init__(self, delta, **kwargs):
        _Indicator.__init__(self, delta=delta, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        delta = params['delta']

        idx_maxs, idx_mins, val_maxs, val_mins = _PeakDetection(delta=delta)(signal)

        if len(idx_maxs) == 0:
            cls.warn("No peak found")
            return _np.nan
        else:
            return _np.nanmax(val_maxs)


class PeaksMin(_Peaks):
    """
    Return the minimum amplitude of detected peaks.

    Parameters
    ----------
    delta : float, >0
        Minimum amplitude of peaks to be selected
    
    Returns
    -------
    mn : float
        Minimum amplitude of detected peaks
    
    """
    def __init__(self, delta, **kwargs):
        _Indicator.__init__(self, delta=delta, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        delta = params['delta']

        idx_maxs, idx_mins, val_maxs, val_mins = _PeakDetection(delta=delta)(data)

        if len(idx_maxs) == 0:
            cls.warn("No peak found, returning numpy.nan")
            return _np.nan
        else:
            return _np.nanmin(val_maxs)


class PeaksMean(_Peaks):
    """
    Return the average amplitude of detected peaks.

    Parameters
    ----------
    delta : float, >0
        Minimum amplitude of peaks to be selected
    
    Returns
    -------
    av : float
        Average amplitude of detected peaks
    
    """
    def __init__(self, delta, **kwargs):
        _Indicator.__init__(self, delta=delta, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        delta = params['delta']

        idx_maxs, idx_mins, val_maxs, val_mins = _PeakDetection(delta=delta)(data)

        if len(idx_maxs) == 0:
            cls.warn("No peak found")
            return _np.nan
        else:
            return _np.nanmean(val_maxs)


class PeaksNum(_Peaks):
    """
    Return the number of detected peaks.

    Parameters
    ----------
    delta : float, >0
        Minimum amplitude of peaks to be selected
    
    Returns
    -------
    n : float
        Number of detected peaks
    
    """
    def __init__(self, delta, **kwargs):
        _Indicator.__init__(self, delta=delta, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        delta = params['delta']

        idx_maxs, idx_mins, val_maxs, val_mins = _PeakDetection(delta=delta)(signal)

        if len(idx_maxs) == 0:
            cls.warn("No peak found")
            return _np.nan
        else:
            return len(idx_maxs)


class _PeaksInterval(_Peaks):
    """
    Peaks base class
    """
    __metaclass__ = _ABCMeta

    def __init__(self, delta, win_pre=1, win_post=1, **kwargs):
        assert win_pre > 0, 'win_pre must be > 0'
        assert win_post > 0, 'win_post must be > 0'
        _Indicator.__init__(self, delta=delta, win_pre=win_pre, win_post=win_post, **kwargs)

    @classmethod
    @_abstract
    def algorithm(cls, data, params):
        pass


class DurationMin(_PeaksInterval):
    """
    Return the minimum duration of detected peaks.

    Parameters
    ----------
    delta : float, >0
        Minimum amplitude of peaks to be selected
    win_pre : float, >0, default=1
        Interval before a detected peak where to search the start of the peak
    win_post : float, >0, default=1
        Interval after a detected peak where to search the end of the peak
        
    Returns
    -------
    mn : float
        Minimum duration of detected peaks
    
    """
    def __init__(self, delta, win_pre=1, win_post=1, **kwargs):
        _Indicator.__init__(self, delta=delta, win_pre=win_pre, win_post=win_post, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        delta = params['delta']
        win_pre = params['win_pre']
        win_post = params['win_post']

        idx_maxs, idx_mins, val_maxs, val_mins = _PeakDetection(delta=delta)(signal)
        if len(idx_maxs) == 0:
            cls.warn("No peaks found")
            return _np.nan

        idxs_start, idxs_stop = _PeakSelection(indices=idx_maxs, win_pre=win_pre, win_post=win_post)(signal)

        if len(idxs_start) == 0:
            cls.warn("Unable to detect the start of the peaks")
            return _np.nan
        else:
            durations = _Durations(starts=idxs_start, stops=idxs_stop)(signal)
            return _np.nanmin(_np.array(durations))


class DurationMax(_PeaksInterval):
    """
    Return the maximum duration of detected peaks.

    Parameters
    ----------
    delta : float, >0
        Minimum amplitude of peaks to be selected
    win_pre : float, >0, default=1
        Interval before a detected peak where to search the start of the peak
    win_post : float, >0, default=1
        Interval after a detected peak where to search the end of the peak
        
    Returns
    -------
    mx : float
        Maximum duration of detected peaks
    
    """
    def __init__(self, delta, win_pre=1, win_post=1, **kwargs):
        _Indicator.__init__(self, delta=delta, win_pre=win_pre, win_post=win_post, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):

        delta = params['delta']
        win_pre = params['win_pre']
        win_post = params['win_post']

        idx_maxs, idx_mins, val_maxs, val_mins = _PeakDetection(delta=delta)(signal)
        if len(idx_maxs) == 0:
            cls.warn("No peaks found")
            return _np.nan

        idxs_start, idxs_stop = _PeakSelection(indices=idx_maxs, win_pre=win_pre, win_post=win_post)(signal)

        if len(idxs_start) == 0:
            cls.warn("Unable to detect the start of the peaks")
            return _np.nan
        else:
            durations = _Durations(starts=idxs_start, stops=idxs_stop)(signal)
            return _np.nanmax(_np.array(durations))


class DurationMean(_PeaksInterval):
    """
    Return the average duration of detected peaks.

    Parameters
    ----------
    delta : float, >0
        Minimum amplitude of peaks to be selected
    win_pre : float, >0, default=1
        Interval before a detected peak where to search the start of the peak
    win_post : float, >0, default=1
        Interval after a detected peak where to search the end of the peak
        
    Returns
    -------
    av : float
        Average duration of detected peaks
    
    """
    def __init__(self, delta, win_pre=1, win_post=1, **kwargs):
        _Indicator.__init__(self, delta=delta, win_pre=win_pre, win_post=win_post, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        delta = params['delta']
        win_pre = params['win_pre']
        win_post = params['win_post']

        idx_maxs, idx_mins, val_maxs, val_mins = _PeakDetection(delta=delta)(signal)
        if len(idx_maxs) == 0:
            cls.warn("No peaks found")
            return _np.nan

        idxs_start, idxs_stop = _PeakSelection(indices=idx_maxs, win_pre=win_pre, win_post=win_post)(signal)

        if len(idxs_start) == 0:
            cls.warn("Unable to detect the start of the peaks")
            return _np.nan
        else:
            durations = _Durations(starts=idxs_start, stops=idxs_stop)(signal)
            return _np.nanmean(_np.array(durations))


class SlopeMin(_PeaksInterval):
    """
    Return the minimum slope of detected peaks.

    Parameters
    ----------
    delta : float, >0
        Minimum amplitude of peaks to be selected
    win_pre : float, >0, default=1
        Interval before a detected peak where to search the start of the peak
    win_post : float, >0, default=1
        Interval after a detected peak where to search the end of the peak
        
    Returns
    -------
    mn : float
        Minimum slope of detected peaks
    
    """
    def __init__(self, delta, win_pre=1, win_post=1, **kwargs):
        _Indicator.__init__(self, delta=delta, win_pre=win_pre, win_post=win_post, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        delta = params['delta']
        win_pre = params['win_pre']
        win_post = params['win_post']

        idx_maxs, idx_mins, val_maxs, val_mins = _PeakDetection(delta=delta)(signal)
        if len(idx_maxs) == 0:
            cls.warn("No peaks found")
            return _np.nan

        idxs_start, idxs_stop = _PeakSelection(indices=idx_maxs, win_pre=win_pre, win_post=win_post)(signal)
        
        if len(idxs_start) == 0:
            cls.warn("Unable to detect the start of the peaks")
            return _np.nan
        else:
            slopes = _Slopes(starts=idxs_start, peaks=idx_maxs)(signal)
            return _np.nanmin(_np.array(slopes))


class SlopeMax(_PeaksInterval):
    """
    Return the maximum slope of detected peaks.

    Parameters
    ----------
    delta : float, >0
        Minimum amplitude of peaks to be selected
    win_pre : float, >0, default=1
        Interval before a detected peak where to search the start of the peak
    win_post : float, >0, default=1
        Interval after a detected peak where to search the end of the peak
        
    Returns
    -------
    mx : float
        Maximum slope of detected peaks
    
    """
    def __init__(self, delta, win_pre=1, win_post=1, **kwargs):
        _Indicator.__init__(self, delta=delta, win_pre=win_pre, win_post=win_post, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        delta = params['delta']
        win_pre = params['win_pre']
        win_post = params['win_post']

        idx_maxs, idx_mins, val_maxs, val_mins = _PeakDetection(delta=delta)(signal)
        if len(idx_maxs) == 0:
            cls.warn("No peaks found")
            return _np.nan

        idxs_start, idxs_stop = _PeakSelection(indices=idx_maxs, win_pre=win_pre, win_post=win_post)(signal)
        
        if len(idxs_start) == 0:
            cls.warn("Unable to detect the start of the peaks")
            return _np.nan
        else:
            slopes = _Slopes(starts=idxs_start, peaks=idx_maxs)(signal)
            return _np.nanmax(_np.array(slopes))


class SlopeMean(_PeaksInterval):
    """
    Return the average slope of detected peaks.

    Parameters
    ----------
    delta : float, >0
        Minimum amplitude of peaks to be selected
    win_pre : float, >0, default=1
        Interval before a detected peak where to search the start of the peak
    win_post : float, >0, default=1
        Interval after a detected peak where to search the end of the peak
        
    Returns
    -------
    mx : float
        Maximum slope of detected peaks
    
    """
    def __init__(self, delta, win_pre=1, win_post=1, **kwargs):
        _Indicator.__init__(self, delta=delta, win_pre=win_pre, win_post=win_post, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        delta = params['delta']
        win_pre = params['win_pre']
        win_post = params['win_post']

        idx_maxs, idx_mins, val_maxs, val_mins = _PeakDetection(delta=delta)(signal)
        if len(idx_maxs) == 0:
            cls.warn("No peaks found")
            return _np.nan

        idxs_start, idxs_stop = _PeakSelection(indices=idx_maxs, win_pre=win_pre, win_post=win_post)(signal)
        if len(idxs_start) == 0:
            cls.warn("Unable to detect the start of the peaks")
            return _np.nan
        else:
            slopes = _Slopes(starts=idxs_start, peaks=idx_maxs)(signal)
            return _np.nanmean(_np.array(slopes))



# coding=utf-8
import numpy as np
from scipy import interpolate
__author__ = 'AleB'


class AbstractCalledError(RuntimeError):
    pass


def abstractmethod(funcobj):
    """A decorator indicating abstract methods.
    """
    def abstract_error():
        #
        #
        #
        #
        #
        raise AbstractCalledError("This method is abstract")
        #
        #
        #
        #
        #

    if hasattr(abstract_error, "func_code"):
        funcobj._func_code = abstract_error.func_code
    else:
        funcobj.__code__ = abstract_error.__code__
    return abstract_error


def derive(data, labels):
    ll = []
    tt = []
    ii = []
    ts = 0
    pre = None
    for i in range(len(labels)):
        if pre != labels[i]:
            ll.append(labels[i])
            tt.append(ts)
            ii.append(i)
            ts += data[i]
            pre = labels[i]
    return ll, tt, ii


def power(spec, freq, min_freq, max_freq):
    """
    Returns the power calculated in the specified band of the spec-freq spectrum
    :param max_freq:
    :param min_freq:
    :param freq:
    :param spec:
    """
    band = np.array([spec[i] for i in range(len(spec)) if min_freq <= freq[i] < max_freq])
    return np.sum(band) / len(spec)


def interpolate_ibi(rr, interp_freq):
    """
    Returns as a tuple the interpolated RR and BT arrays
    :param interp_freq:
    :param rr:
    """
    step = 1.0 / interp_freq
    rr /= 1000
    rr = np.array(rr)
    bt = np.cumsum(rr)
    x_min = bt[0]
    x_max = bt[-1]
    bt = np.insert(bt, 0, 0)
    bt = np.append(bt, bt[-1] + 1)
    rr = np.insert(rr, 0, 0)
    rr = np.append(rr, rr[-1])
    tck = interpolate.splrep(bt, rr)
    bt_interp = np.arange(x_min, x_max, step)
    rr_interp = interpolate.splev(bt_interp, tck)
    return rr_interp, bt_interp


def template_interpolation(x, t, step, template=None):
    if template is None:
        template = np.square(np.cos(np.arange(0, 0.505, 0.005) * np.pi))

    x_old = x[0]
    t_old = t[0]

    x_out = np.array([])
    t_out = np.array([])

    for i in range(1, len(x)):
        x_curr = x[i]
        t_curr = t[i]

        x_template = template * (x_old - x_curr) + x_curr
        t_template = np.linspace(t_old, t_curr, 101)

        x_out = np.hstack((x_out, x_template, x_curr))
        t_out = np.hstack((t_out, t_template, t_curr))

        t_old = t_curr
        x_old = x_curr

    t_output = np.arange(t[0], t[-1], step)

    f = interpolate.interp1d(t_out, x_out)
    x_output = f(t_output)
    return x_output, t_output


class PhUI(object):
    @staticmethod
    def a(condition, message):
        if not condition:
            raise ValueError(message)

    @staticmethod
    def o(mex):
        PhUI.p(mex, '', 37)

    @staticmethod
    def i(mex):
        PhUI.p(mex, '', 94)

    @staticmethod
    def w(mex):
        PhUI.p(mex, 'Warning: ', 33)

    @staticmethod
    def e(mex):
        PhUI.p(mex, 'Error: ', 31)

    @staticmethod
    def p(mex, lev, col):
        print(">%s\x1b[%dm%s\x1b[39m" % (lev, col, mex))



# coding=utf-8
from __future__ import division
from . import *


def test_ibi_extraction():
    # %%
    FSAMP = 2048
    TSTART = 0

    # %%
    # TEST IBI EXTRACTION FROM ECG
    ecg = ph.EvenlySignal(TestData.ecg(), sampling_freq=FSAMP, signal_nature='ECG', start_time=TSTART)

    ecg = ecg.resample(fout=4096, kind='cubic')

    ibi = ph.BeatFromECG()(ecg)
    assert (int(np.mean(ibi) * 10000) == 8619)

    # %% Test ibi correction
    id_bad_ibi = ph.BeatOutliers(cache=3, sensitivity=0.25)(ibi)
    assert (len(id_bad_ibi) == 0)

    def insert_ibi(id_ibi, ibi_good):
        t_ibi_bad = ibi_good.get_times()
        v_ibi_bad = ibi_good.get_values()
        t_ibi_bad_left = t_ibi_bad[id_ibi]
        t_ibi_bad_right = t_ibi_bad[id_ibi + 1]
        t_ibi_add = t_ibi_bad_left + (t_ibi_bad_right - t_ibi_bad_left) / 4

        t_ibi_bad = np.r_[t_ibi_bad[:id_ibi + 1], t_ibi_add, t_ibi_bad[id_ibi + 1:]]

        v_ibi_add = t_ibi_add - t_ibi_bad_left
        v_ibi_bad[id_ibi + 1] = t_ibi_bad_right - t_ibi_add
        v_ibi_bad = np.r_[v_ibi_bad[:id_ibi + 1], v_ibi_add, v_ibi_bad[id_ibi + 1:]]

        return (ph.UnevenlySignal(v_ibi_bad, sampling_freq=ibi_good.get_sampling_freq(),
                                  start_time=ibi_good.get_start_time(),
                                  signal_nature=ibi_good.get_signal_nature(),
                                  x_values=t_ibi_bad, x_type='instants'))

    ibi_bad = insert_ibi(20, ibi.copy())
    id_bad_ibi = ph.BeatOutliers(cache=3, sensitivity=0.25)(ibi_bad)

    assert (id_bad_ibi[0] == 21 and id_bad_ibi[1] == 22)

    # %%
    # TEST IBI EXTRACTION FROM BVP
    bvp = ph.EvenlySignal(TestData.bvp(), sampling_freq=FSAMP, signal_nature='BVP', start_time=TSTART)
    bvp = bvp.resample(fout=4096, kind='cubic')

    ibi = ph.BeatFromBP()(bvp)

    id_bad_ibi = ph.BeatOutliers(cache=3, sensitivity=0.25)(ibi)
    assert (np.sum(id_bad_ibi) == 744)



# coding=utf-8
from __future__ import division

import unittest
from . import ph, TestData, np

__author__ = 'aleb'


class FiltersTest(unittest.TestCase):
    def test_general(self):
        # %%
        FSAMP = 2048
        TSTART = 0

        # %%
        ecg = ph.EvenlySignal(TestData.ecg(), sampling_freq=FSAMP, signal_nature='ECG', start_time=TSTART)

        # %%
        # TEST Normalize
        self.assertAlmostEqual(np.mean(ph.Normalize(norm_method='standard')(ecg)), 0)

        self.assertEqual(np.min(ph.Normalize(norm_method='min')(ecg)), 0)

        self.assertEqual(np.min(ph.Normalize(norm_method='maxmin')(ecg)), 0)
        self.assertEqual(np.max(ph.Normalize(norm_method='maxmin')(ecg)), 1)

        ecg_ = ph.Normalize(norm_method='custom', norm_bias=4, norm_range=0.1)(ecg)
        self.assertAlmostEqual(np.max(ecg_) - np.min(ecg_), 16.41, delta=0.005)

        # %%
        # TEST Diff
        s = ph.EvenlySignal(np.arange(1000), sampling_freq=FSAMP, start_time=TSTART)
        # degree = [5, 50, -1]
        degree = 5
        self.assertAlmostEqual(np.mean(ph.Diff(degree=degree)(s)), degree)  # OK

        # %% TEST IIRFilter
        self.assertAlmostEqual(int(np.max(ph.IIRFilter(fp=10, fs=70)(ecg)) * 10000), 8238)
        self.assertAlmostEqual(int(np.max(ph.IIRFilter(fp=70, fs=45)(ecg)) * 10000), 2144)
        self.assertAlmostEqual(int(np.max(ph.IIRFilter(fp=[5, 25], fs=[0.05, 50], ftype='ellip')(ecg)) * 10000), 8786)

        # %%

        # TEST ConvolutionalFilter
        self.assertAlmostEqual(np.max(ph.ConvolutionalFilter(irftype='gauss', win_len=0.1)(ecg)), .7501, delta=.0001)
        self.assertAlmostEqual(np.max(ph.ConvolutionalFilter(irftype='rect', win_len=0.1)(ecg)), .4022, delta=.0001)
        self.assertAlmostEqual(np.max(ph.ConvolutionalFilter(irftype='triang', win_len=0.1)(ecg)), .5000, delta=.0001)
        self.assertAlmostEqual(
            np.max(ph.ConvolutionalFilter(irftype='dgauss', win_len=0.1, normalize=False)(ecg)), .8823, delta=.0001)

        irf = np.r_[np.arange(50, 500, 5)]
        ecg_cfC = ph.ConvolutionalFilter(irftype='custom', irf=irf, normalize=True)(ecg)
        self.assertEqual(int(np.max(ecg_cfC) * 10000), 7723)

        # %%
        # TEST DeConvolutionalFilter
        ecg_df = ph.DeConvolutionalFilter(irf=irf, normalize=True, deconv_method='fft')(ecg_cfC)  # OK
        self.assertEqual(int(np.max(ecg_df) * 10000), 48973)



# coding=utf-8
from __future__ import division

from numpy import array as _array
from .tools.Tools import *
import numpy as _np
from .filters import Filters
from .segmentation import SegmentsGenerators
from .indicators import FrequencyDomain
from .indicators import NonLinearDomain
from .indicators import PeaksDescription
from .indicators import TimeDomain
from .BaseSegmentation import Segment
from .Signal import EvenlySignal, UnevenlySignal, MultiEvenly, from_pickle, from_pickleable
from .interactive import Annotate
# BE CAREFUL with NAMES!!!
from .estimators.Estimators import *
from .filters.Filters import *
from .indicators.FrequencyDomain import *
from .indicators.NonLinearDomain import *
from .indicators.PeaksDescription import *
from .indicators.TimeDomain import *
from .sqi.SignalQuality import *
#from .tests import TestData
from .segmentation.SegmentsGenerators import *

#TODO: all signals as N_SAMPLES x N_CH, with N_CH =1 for non MultiEvenly

print("Please cite:")
print("Bizzego et al. (2019) 'pyphysio: A physiological signal processing library for data science approaches in physiology', SoftwareX")

__author__ = "AleB"

def nature2type(data):
    data.ph['signal_type'] = data.ph['signal_nature']
    if isinstance(data, NIRS):
        stim = data.get_stim()
        stim.ph['signal_type'] = stim.ph['signal_nature']
        data.set_stim(stim)
    return(data)
    
def preset_sqi_ecg(prefix="SQI_", method='ar'):
    K = Kurtosis(name='kurtosis')
    SPR = SpectralPowerRatio(method, name='SPR')
    DE = DerivativeEnergy(name='DE')
    
    t = [K, SPR, DE]

    if prefix is not None:
        for i in t:
            i.set(name=prefix + i.get("name"))

    return t

def preset_hrv_fd(prefix="IBI_", method='ar'):
    VLF = PowerInBand(interp_freq=4, freq_max=0.04, freq_min=0.00001, method=method, name="VLF_Pow")
    LF = PowerInBand(interp_freq=4, freq_max=0.15, freq_min=0.04, method=method, name="LF_Pow")
    HF = PowerInBand(interp_freq=4, freq_max=0.4, freq_min=0.15, method=method, name="HF_Pow")
    Total = PowerInBand(interp_freq=4, freq_max=2, freq_min=0.00001, method=method, name="Total_Pow")
    
    t = [VLF, LF, HF, Total]

    if prefix is not None:
        for i in t:
            i.set(name=prefix + i.get("name"))

    return t


def preset_hrv_td(prefix="IBI_"):
    rmssd = RMSSD(name="RMSSD")
    sdsd = SDSD(name="SDSD")
    RRmean = Mean(name="Mean")
    RRstd = StDev(name="RRstd")
    RRmedian = Median(name="Median")
    pnn10 = PNNx(threshold=10, name="pnn10")
    pnn25 = PNNx(threshold=25, name="pnn25")
    pnn50 = PNNx(threshold=50, name="pnn50")
    mn = Min(name="Min")
    mx = Max(name="Max")
    sd1 = PoincareSD1(name="sd1")
    sd2 = PoincareSD2(name="sd2")
    sd12 = PoincareSD1SD2(name="sd12")
    sdell = PoinEll(name="sdell")
    DFA1 = DFAShortTerm(name="DFA1")
    DFA2 = DFALongTerm(name="DFA2")

    t = [rmssd, sdsd, RRmean, RRstd, RRmedian, pnn10, pnn25, pnn50, mn, mx, sd1, sd2, sd12,
         sdell, DFA1, DFA2]

    if prefix is not None:
        for i in t:
            i.set(name=prefix + i.get("name"))

    return t


def preset_phasic(delta, prefix="pha_"):
    mean = Mean()
    std = StDev()
    rng = Range()
    pks_max = PeaksMax(delta=delta)
    pks_min = PeaksMin(delta=delta)
    pks_mean = PeaksMean(delta=delta)
    n_peaks = PeaksNum(delta=delta)
    dur_mean = DurationMean(delta=delta, win_pre=2, win_post=2)
    slopes_mean = SlopeMean(delta=delta, win_pre=2, win_post=2)
    auc = AUC()

    t = [mean, std, rng, pks_max, pks_min, pks_mean, n_peaks, dur_mean, slopes_mean, auc]

    if prefix is not None:
        for i in t:
            i.set(name=prefix + i.__class__.__name__)

    return t


def preset_tonic(prefix="ton_"):
    mean = Mean()
    std = StDev()
    rng = Range()
    auc = AUC()

    t = [mean, std, rng, auc]

    if prefix is not None:
        for i in t:
            i.set(name=prefix + i.__class__.__name__)

    return t

def preset_eeg(prefix="eeg_", method='welch'):
    delta = PowerInBand(freq_min=0, freq_max=3, method=method, name="delta")
    theta = PowerInBand(freq_min=3.5, freq_max=7.5, method=method, name="theta")
    alpha = PowerInBand(freq_min=7.5, freq_max=13, method=method, name="alpha")
    beta = PowerInBand(freq_min=14, freq_max=30, method=method, name="beta")
    total = PowerInBand(freq_min=0, freq_max=100, method=method, name="total")
    
    t = [delta, theta, alpha, beta, total]

    if prefix is not None:
        for i in t:
            i.set(name=prefix + i.get("name"))

    return t


def preset_emg(prefix='emg_', method = 'welch'):
    mx = Max(name='maximum')
    mn = Min(name='minimum')
    mean = Mean(name='mean')
    rng = Range(name='range')
    sd = StDev(name='sd')
    auc = AUC(name='auc')
    en4_40 = PowerInBand(freq_min=4, freq_max=40, method=method, name="en_4_40")
    
    t = [mx, mn, mean, rng, sd, auc, en4_40]

    if prefix is not None:
        for i in t:
            i.set(name=prefix + i.get("name"))

    return t


def preset_resp(prefix='resp', method='welch'):
    e_low = PowerInBand(freq_min=0, freq_max=0.25, method=method, name="energy_low")
    e_high = PowerInBand(freq_min=0.25, freq_max=5, method=method, name="energy_high")
    resp_rate = PeakInBand(freq_min=0.25, freq_max=5, method=method, name="resp_rate")
    
    t = [e_low, e_high, resp_rate]
    
    if prefix is not None:
        for i in t:
            i.set(name=prefix + i.get("name"))

    return t

def preset_activity(prefix='activity', method='welch'):
    mx = Max(name='maximum')
    mn = Min(name='minimum')
    mean = Mean(name='mean')
    rng = Range(name='range')
    sd = StDev(name='sd')
    auc = AUC(name='auc')
    en_25 = PowerInBand(freq_min=0, freq_max=25, method=method, name="en_25")
    
    t = [mx, mn, mean, rng, sd, auc, en_25]

    if prefix is not None:
        for i in t:
            i.set(name=prefix + i.get("name"))

    return t
 
def fmap(segments, algorithms, alt_signal=None):
    # TODO : rename extract_indicators
    """
    Generates a list composed of a list of results for each segment.

    [[result for each algorithm] for each segment]
    :param segments: An iterable of segments (e.g. an initialized SegmentGenerator)
    :param algorithms: A list of algorithms
    :param alt_signal: The signal that will be used instead of the one referenced in the segments

    :return: values, col_names A tuple: matrix (segment x algorithms) containing a value for each
     algorithm, the list of the algorithm names.
    """
    from numpy import asarray as _asarray
    
    seg_for = segments(alt_signal) if isinstance(segments, SegmentsGenerator) else segments
    
    values = []
    for seg in seg_for:
        segment_data = _np.array([seg.get_begin_time(), seg.get_end_time(), seg.get_label()]).reshape(3,1)
        vals_segment = []
        for alg in algorithms:
            vals_alg = _np.array(alg(seg(alt_signal)))

            if not alt_signal.is_multi():
#                vals_alg = _np.expand_dims([vals_alg], 1)
                vals_alg = _np.array([vals_alg])
            vals_segment.append(vals_alg)
            
        vals_segment = _np.array(vals_segment)
        seg_data_array = _np.repeat(segment_data, alt_signal.get_nchannels(), axis = 1)
        vals_segment = _np.concatenate([seg_data_array, vals_segment], axis = 0)
        values.append(vals_segment)
    
    values = _np.array(values)
    
    #for compatibility
    if not alt_signal.is_multi():
        values = values[:,:,0]
        
    col_names = ["begin", "end", "label"] + [x.__repr__() for x in algorithms]
    
    return values, _array(col_names)


def algo(function, **kwargs):
    """
    Builds on the fly a new algorithm class using the passed function and params if passed.
    Note that the cache will not be available.
    :param function: function(data, params) to be called
    :param kwargs: parameters to pass to the function.
    :return: An algorithm class if params is None else a parametrized algorithm instance.
    """

    from .BaseAlgorithm import Algorithm

    class Custom(Algorithm):
        def __init__(self, **kwargs):
            Algorithm.__init__(self, **kwargs)

        @classmethod
        def algorithm(cls, data, params):
            return function(data, params)

    if len(kwargs) == 0:
        return Custom
    else:
        return Custom(**kwargs)


def test():
    from pytest import main as m
    from os.path import dirname as d
    m(['-x', d(__file__)])



# coding=utf-8
from __future__ import print_function
from __future__ import division

import matplotlib.pyplot as plt
import pyphysio as ph
import numpy as np


class _MouseSelectionFilter(object):
    def __init__(self, onselect):
        self._select = onselect
        self._last_press = None

    def on_move(self, event):
        self._last_press = None

    def on_press(self, event):
        x, y = event.xdata, event.ydata
        self._last_press = x, y, event.button

    def on_release(self, event):
        x, y = event.xdata, event.ydata
        if self._last_press is not None:
            xx, yy, b = self._last_press
            if x == xx and y == yy and event.button == b:
                self._select(event)


class _ItemManager(object):
    def __init__(self, snap_func, select, unselect, add, delete):
        self._snap_func = snap_func
        self._select = select
        self._unselect = unselect
        self._delete = delete
        self._add = add
        self.selection = -1

    def unselect(self):
        self._unselect(self.selection)
        self.selection = None

    def on_select(self, ev):
        if ev.xdata is not None and ev.ydata is not None:
            x, y, item, new = self._snap_func(ev.xdata, ev.ydata)
#            print("on_select: %d, %d: %d" % (x, y, item))
            if self.selection is not None:
                self.unselect()
            if ev.button == 1:
                if new:
                    self._add(x, y, item)
                else:
                    self.selection = item
                    self._select(item)


class Annotate(object):
    def __init__(self, ecg, ibi):
        self.plots = None
        self.peaks_t = None
        self.done = False
        self.ecg = ecg
        self.ibi = ibi
        self.fig = plt.figure()
        self.p_sig = self.fig.add_subplot(2, 1, 1)
        self.p_res = self.fig.add_subplot(2, 1, 2, sharex=self.p_sig)

        self.max = ph.Max()(self.ecg)
        self.min = ph.Min()(self.ecg)

        self.margin = ph.Range()(self.ecg) * .1
        self.max += self.margin
        self.min -= self.margin

        if isinstance(ibi, ph.UnevenlySignal):
            self.peaks_t = self.ibi.get_times()
            self.peaks_v = self.ibi.get_values()
        else:
            self.peaks_t = np.empty(0)
            self.peaks_v = np.empty(0)

        self.p_sig.plot(self.ecg.get_times(), self.ecg.get_values(), 'b')

        self.p_res.plot(self.peaks_t, self.peaks_v, 'b'),
        self.p_res.plot(self.peaks_t, self.peaks_v, 'go')

        self.replot()

        class Cursor(object):
            left = None
            right = None
            radius = .3
            radiusi = int(radius * self.ecg.get_sampling_freq())

            @staticmethod
            def on_move(event):
                Cursor.draw(event)

            @staticmethod
            def on_scroll(event):
                if event.button == "up":
                    Cursor.radiusi += 3
                elif event.button == "down":
                    Cursor.radiusi -= 7
                Cursor.radius = Cursor.radiusi / self.ecg.get_sampling_freq()
                Cursor.draw(event)

            @staticmethod
            def draw(event):
                if Cursor.left is not None:
                    Cursor.left.remove()
                    Cursor.right.remove()
                    Cursor.left = None
                    Cursor.right = None
                if event.xdata is not None:  # TODO (Andrea): not do this if speed (dxdata/dt) is high
                    Cursor.left = self.p_sig.vlines(event.xdata - Cursor.radius, self.min - self.margin * 2,
                                                    self.max + self.margin * 2, 'k')
                    Cursor.right = self.p_sig.vlines(event.xdata + Cursor.radius, self.min - self.margin * 2,
                                                     self.max + self.margin * 2, 'k')
                self.fig.canvas.draw()

        def find_peak(s):
            return np.argmax(s)

        def snap(xdata, ydata):
            nearest_after = self.peaks_t.searchsorted(xdata)
            nearest_prev = nearest_after - 1

            dist_after = self.peaks_t[nearest_after] - xdata if 0 <= nearest_after < len(self.peaks_t) else None
            dist_prev = xdata - self.peaks_t[nearest_prev] if 0 <= nearest_prev < len(self.peaks_t) else None

            if dist_after is None or dist_prev < dist_after:
                if dist_prev is not None and dist_prev < Cursor.radius:
                    return self.peaks_t[nearest_prev], ydata, nearest_prev, False
            elif dist_prev is None or dist_after < dist_prev:
                if dist_after is not None and dist_after < Cursor.radius:
                    return self.peaks_t[nearest_after], ydata, nearest_after, False

            s = self.ecg.segment_time(xdata - Cursor.radius, xdata + Cursor.radius)
            s = np.array(s)
            m = find_peak(s)
            return xdata - Cursor.radius + m / self.ecg.get_sampling_freq(), ydata, nearest_after, True

        class Selector(object):
            selector = None

            @staticmethod
            def select(item):
#                print("select: %d" % item)
                Selector.selector = self.p_sig.vlines(self.peaks_t[item], self.min - self.margin, self.max + self.margin, 'g')

            @staticmethod
            def unselect(item):
                if Selector.selector is not None:
#                    print("unselect: %d" % item)
                    Selector.selector.remove()

        # it is correct that the computation of the values is done at the end (line 186)
        def add(time, y, pos):
            self.peaks_t = np.insert(self.peaks_t, pos, time)
            self.replot()

        def delete(item):
            self.peaks_t = np.delete(self.peaks_t, item)
            self.replot()

        im = _ItemManager(snap, Selector.select, Selector.unselect, add, delete)
        mf = _MouseSelectionFilter(im.on_select)

        def press(ev):
#            print(ev.key)
            if ev.key == "d" and im.selection is not None:
                delete(im.selection)
                im.unselect()
                
        def handle_close(ev):
            self.done = True
            return
                

            
        clim = self.fig.canvas.mpl_connect('motion_notify_event', lambda e: (mf.on_move(e), Cursor.on_move(e)))
        clip = self.fig.canvas.mpl_connect('button_press_event', mf.on_press)
        clir = self.fig.canvas.mpl_connect('button_release_event', mf.on_release)
        clis = self.fig.canvas.mpl_connect('scroll_event', Cursor.on_scroll)
        clik = self.fig.canvas.mpl_connect('key_press_event', press)
        ccls = self.fig.canvas.mpl_connect('close_event', handle_close)
        
        while not self.done :
#            print('waiting')
            plt.pause(1)
        
        plt.close(self.fig)
        # it is correct that the computation of the values is done at the end!
        # do not change!
        self.peaks_v = np.diff(self.peaks_t)
        self.peaks_v = np.r_[self.peaks_v[0], self.peaks_v]
                    
        if isinstance(ibi, ph.UnevenlySignal):
            self.ibi_ok =  ph.UnevenlySignal(values=self.peaks_v,
                                     sampling_freq=self.ibi.get_sampling_freq(),
                                     signal_type=self.ibi.get_signal_type(),
                                     start_time=self.ibi.get_start_time(),
                                     x_values=self.peaks_t,
                                     x_type='instants',
                                     duration=self.ibi.get_duration())
        else:
            self.ibi_ok = ph.UnevenlySignal(values=self.peaks_v,
                                     sampling_freq=self.ecg.get_sampling_freq(),
                                     signal_type=self.ecg.get_signal_type(),
                                     start_time=self.ecg.get_start_time(),
                                     x_values=self.peaks_t,
                                     x_type='instants',
                                     duration=self.ecg.get_duration())
        
    def __call__(self):
        return self.ibi_ok
    
    def replot(self):
        if self.plots is not None:
            self.plots.remove()
        if self.peaks_t is not None:
            self.plots = self.p_sig.vlines(self.peaks_t, self.min, self.max, 'y')
            self.fig.canvas.draw()



# coding=utf-8
from __future__ import division

from . import ph, TestData, approx
import numpy as np

import pytest

__author__ = 'aleb'


# noinspection PyAttributeOutsideInit
class TestSignal(object):
    def setup_class(self):
        self.samples = 1000
        self.freq_down = 256
        self.freq1 = 100
        self.freq2 = 1024
        self.start1 = 3
        self.start2 = 13370003
        self.nature = "una_bif-fa"

        np.random.seed(1234)
        self.x_vals = np.cumsum(np.random.rand(1, self.samples) * 9 + 1).astype(int)
        self.y_vals = np.cumsum(np.random.rand(1, self.samples) - .5) * 100

        self.s = ph.EvenlySignal(values=TestData.ecg(),
                                 sampling_freq=self.freq1,
                                 signal_nature=self.nature,
                                 start_time=self.start1,
                                 )

        self.empty_s = ph.EvenlySignal(values=[],
                                       sampling_freq=self.freq1,
                                       signal_nature=self.nature,
                                       start_time=self.start1,
                                       )

        self.us = ph.UnevenlySignal(values=self.y_vals,
                                    x_values=self.x_vals,
                                    sampling_freq=self.freq1,
                                    signal_nature=self.nature,
                                    start_time=self.start1,
                                    x_type='indices'
                                    )

    def test_signal_instantiation(self):
        # instants
        ust = ph.UnevenlySignal(values=self.y_vals,
                                x_values=self.x_vals / self.freq1 + self.start1,
                                sampling_freq=self.freq1,
                                signal_nature=self.nature,
                                start_time=self.start1,
                                x_type='instants'
                                )

        assert len(ust) == len(self.us)
        assert len(ust.get_indices()) == len(self.x_vals)  # FIXME early tests for get_indices()

        for i, (x, y) in enumerate(zip(ust.get_indices(), self.x_vals)):
            assert x == y, "index %d" % i

        # start_time on first time
        ph.UnevenlySignal([1, 2, 3],
                          x_values=[1, 2, 3],
                          x_type='instants',
                          sampling_freq=10,
                          start_time=1,
                          )

        # start_time negative
        ph.UnevenlySignal([1, 2, 3],
                          x_values=[1, 2, 3],
                          x_type='instants',
                          sampling_freq=10,
                          start_time=-1.1,
                          )

        # start_time after first time
        with pytest.raises(AssertionError):
            ph.UnevenlySignal([1, 2, 3],
                              x_values=[1, 2, 3],
                              x_type='instants',
                              sampling_freq=10,
                              start_time=1.1,
                              )

        # start_time not numeric
        with pytest.raises(AssertionError):
            ph.EvenlySignal([], 1024, start_time='ugo')

        # None values
        with pytest.raises(AssertionError):
            ph.EvenlySignal(None, 1024)

        # non 1-dim values 1
        with pytest.raises(AssertionError):
            ph.EvenlySignal([[1], [2]], 1024)

        # non 1-dim values 2
        with pytest.raises(AssertionError):
            ph.EvenlySignal([[1, 2]], 1024)

        # x_values missing
        with pytest.raises(AssertionError):
            ph.UnevenlySignal([])

        # x_values not strictly monotonic
        with pytest.raises(AssertionError):
            ph.UnevenlySignal(range(3), x_values=[0, 1.0, 1])

        # x_values length mismatch
        with pytest.raises(AssertionError):
            ph.UnevenlySignal(range(3), x_values=range(4))

        # x_type unknown
        with pytest.raises(AssertionError):
            ph.UnevenlySignal(range(3), x_values=range(3), x_type='bart')

    def test_signal_base(self):
        # ph
        assert isinstance(self.s.ph, dict)
        assert len(self.s.ph) >= 3

        # get_start_time
        assert self.s.get_start_time() == self.start1

        # set_start_time
        self.s.set_start_time(self.start2)
        assert self.s.get_start_time() == self.start2

        # get_sampling_freq
        assert self.s.get_sampling_freq() == self.freq1

        # set_sampling_freq
        self.s.set_sampling_freq(self.freq2)
        assert self.s.get_sampling_freq() == self.freq2

        # EvenlySignal#get_end_time
        assert self.s.get_end_time() == self.s.get_start_time() + len(self.s) / self.s.get_sampling_freq()  # future div

        # UnevenlySignal#get_end_time
        assert self.us.get_end_time() == self.us.get_start_time() + (self.x_vals[-1] + 1.) / self.us.get_sampling_freq()

        # get_duration
        assert isinstance(self.s.get_duration(), float)
        assert self.s.get_duration() == self.s.get_end_time() - self.s.get_start_time()

        # get_idx
        assert self.s.get_idx(self.s.get_start_time()) == 0
        assert self.s.get_idx(self.s.get_end_time()) is not None
        assert self.s.get_idx(10) == (10 - self.s.get_start_time()) * self.s.get_sampling_freq()

        # plot
        import matplotlib.lines
        a = self.s.plot()
        assert isinstance(a, list)
        assert isinstance(a[0], matplotlib.lines.Line2D)
        import matplotlib.collections
        a = self.s.plot("ro")
        assert isinstance(a, list)
        assert isinstance(a[0], matplotlib.lines.Line2D)
        assert isinstance(self.s.plot("|r"), matplotlib.collections.LineCollection)

        # pickleability
        ps = ph.Signal.from_pickleable(self.s.pickleable)
        assert ps.get_start_time() == self.s.get_start_time()
        assert ps.get_sampling_freq() == self.s.get_sampling_freq()
        assert ps.get_end_time() == self.s.get_end_time()
        assert ps.get_signal_nature() == self.s.get_signal_nature()
        assert len(ps) == len(self.s)

    def test_evenly_signal_base(self):
        # get_times
        t = self.s.get_times()
        assert len(self.s.get_values()) == len(t)  # length
        assert len(np.where(np.diff(t) <= 0)[0]) == 0  # strong monotonicity

        # get_time
        assert self.s.get_time(0) == self.s.get_start_time()
        assert self.s.get_time(len(t) - 1) == approx(self.s.get_end_time(), rel=.00002)
        assert self.s.get_time(len(t)) is not None
        assert self.s.get_time(10) == self.s.get_start_time() + 10 / self.s.get_sampling_freq()

        # get_iidx
        assert self.s.get_iidx(0) == self.s.get_idx(0)
        assert self.s.get_iidx(len(t) - 1) == self.s.get_idx(len(t) - 1)
        assert self.s.get_iidx(len(t)) == self.s.get_idx(len(t))

        # get_time_from_iidx
        assert self.s.get_time_from_iidx(0) == self.s.get_time(0)
        assert self.s.get_time_from_iidx(len(t) - 1) == self.s.get_time(len(t) - 1)
        assert self.s.get_time_from_iidx(len(t)) == self.s.get_time(len(t))

    def test_unevenly_signal_base(self):
        # get_indices
        for i, (a, b) in enumerate(zip(self.us.get_indices(), self.x_vals)):
            assert a == b

        # get_times
        for i, (a, b) in enumerate(zip(self.us.get_times(),
                                       self.x_vals / self.us.get_sampling_freq() + self.us.get_start_time())):
            assert a == b, "Index %d" % i

        # get_time
        assert self.us.get_time(0) == self.us.get_start_time()
        assert self.us.get_time_from_iidx(len(self.us) - 1) + 1. / self.us.get_sampling_freq() == self.us.get_end_time()
        assert self.us.get_time_from_iidx(len(self.us)) == self.us.get_time_from_iidx(len(self.us) - 1)
        assert self.us.get_time(10) == self.us.get_start_time() + 10 / self.us.get_sampling_freq()

        #



# coding=utf-8
from __future__ import division
import numpy as _np
from ..BaseEstimator import Estimator as _Estimator
from ..Signal import UnevenlySignal as _UnevenlySignal, EvenlySignal as _EvenlySignal
from ..filters.Filters import IIRFilter as _IIRFilter, DeConvolutionalFilter as _DeConvolutionalFilter, \
    ConvolutionalFilter as _ConvolutionalFilter
from ..tools.Tools import SignalRange as _SignalRange, PeakDetection as _PeakDetection, Minima as _Minima, \
    PeakSelection as _PeakSelection, Diff as _Diff

__author__ = 'AleB'


# IBI ESTIMATION
class BeatFromBP(_Estimator):
    """
    Identify the beats in a Blood Pulse (BP) signal and compute the IBIs.
    Optimized to identify the percussion peak.

    Optional parameters
    -------------------
    
    bpm_max : int, (1, 400], default=120
        Maximal expected heart rate (in beats per minute)
    win_pre : float, (0, 1], default=0.25
        Portion (in seconds) to consider before the candidate beat position where to look for the beat
    win_post : float, (0, 1], default=0.05
        Portion (in seconds) to consider after the candidate beat position where to look for the beat


    Returns
    -------
    ibi : UnevenlySignal
        Inter beat interval values at percussion peaks

    Notes
    -----
    Please cite:
        Bizzego, Andrea, and Cesare Furlanello. "DBD-RCO: Derivative Based Detection And Reverse Combinatorial Optimization To Improve Heart Beat Detection For Wearable Devices." bioRxiv (2017): 118943.
    """
    
    def __init__(self, bpm_max=120, win_pre=.25, win_post=.05):
        if not 10 < bpm_max < 400:
            self.warn("Parameter bpm_max out of reasonable range (10, 400)")
        assert 0 < win_pre <= 1, "Window pre peak value should be in (0 and 1]"
        assert 0 < win_post <= 1, "Window post peak value should be in (0 and 1]"
        _Estimator.__init__(self, bpm_max=bpm_max, win_pre=win_pre, win_post=win_post)

    @classmethod
    def algorithm(cls, signal, params):
        fsamp = signal.get_sampling_freq()
        bpm_max = params["bpm_max"]
        win_pre = params["win_pre"] * fsamp
        win_post = params["win_post"] * fsamp

        fmax = bpm_max / 60
        refractory = 1 / fmax

        # STAGE 1 - EXTRACT BEAT POSITION SIGNAL
        # filtering
        signal_f = _IIRFilter(fp=1.2 * fmax, fs=3 * fmax, ftype='ellip')(signal)
        # find range for the adaptive peak detection
        delta = 0.5 * _SignalRange(win_len=1.5 / fmax, win_step=1 / fmax)(signal_f)
        
        #adjust for delta values equal to 0
        idx_delta_zeros = _np.where(delta==0)[0]
        idx_delta_nozeros = _np.where(delta>0)[0]
        delta[idx_delta_zeros] = _np.min(delta[idx_delta_nozeros])
        
        # detection of candidate peaks
        maxp, minp, ignored, ignored = _PeakDetection(delta=delta, refractory=refractory, start_max=True)(signal_f)

        if maxp[0] == 0:
            maxp = maxp[1:]

        # STAGE 2 - IDENTIFY PEAKS using the signal derivative
        # compute the signal derivative
        dxdt = _Diff()(signal)

        true_peaks = []
        # for each candidate peak find the correct peak
        for idx_beat in maxp:
            start_ = int(idx_beat - win_pre)
            if start_ < 0:
                start_ = 0

            stop_ = int(idx_beat + win_post)
            if stop_ > len(dxdt):
                stop_ = -1

            # select portion of derivative where to search
            obs = dxdt[start_:stop_]
            peak_obs = _np.argmax(obs)
            true_obs = dxdt[start_ + peak_obs: stop_]

            # find the 'first minimum' (zero) the derivative (peak)
            idx_mins, mins = _Minima(win_len=0.1, win_step=0.025, method='windowing')(abs(true_obs))

            if len(idx_mins) >= 1:
                peak = idx_mins[0]
                true_peaks.append(start_ + peak_obs + peak + 1)
            else:
                cls.warn('Peak not found; idx_beat: ' + str(idx_beat))
                pass

        # STAGE 3 - FINALIZE computing IBI
        ibi_values = _np.diff(true_peaks) / fsamp
        ibi_values = _np.r_[ibi_values[0], ibi_values]
        idx_ibi = _np.array(true_peaks)

        ibi = _UnevenlySignal(values=ibi_values,
                              sampling_freq=fsamp,
                              start_time=signal.get_start_time(),
                              signal_type='IBI',
                              x_values=idx_ibi,
                              x_type='indices',
                              duration=signal.get_duration())
        return ibi


class BeatFromECG(_Estimator):
    """
    Identify the beats in an ECG signal and compute the IBIs.

    Optional parameters
    -------------------
    
    bpm_max : int, (1, 400], default=120
        Maximal expected heart rate (in beats per minute)
    delta : float, >=0, default=0
        Threshold for the peak detection. By default it is computed from the signal (adaptive thresholding)
    k : float, (0,1), default=0.7
        Ratio at which the signal range is multiplied (when delta = 0)

    Returns
    -------
    ibi : UnevenlySignal
        Inter beat interval values at percussion peaks

    Notes
    -----
        This algorithms looks for maxima in the signal which are followed by values lower than a delta value. 
        The adaptive version estimates the delta value adaptively.
    """

    def __init__(self, bpm_max=120, delta=0, k=0.7):
        if not 10 < bpm_max < 400:
            self.warn("Parameter bpm_max out of reasonable range (10, 400)")
        assert delta >= 0, "Delta value should be positive (or equal to 0 if automatically computed)"
        assert 0 < k < 1, "K coefficient must be in the range (0,1)"
        _Estimator.__init__(self, bpm_max=bpm_max, delta=delta, k=k)

    @classmethod
    def algorithm(cls, signal, params):
        bpm_max, delta, k = params["bpm_max"], params["delta"], params["k"]
        fmax = bpm_max / 60

        if delta == 0:
            delta = k * _SignalRange(win_len=2 / fmax, win_step=0.5 / fmax, smooth=False)(signal)

        #adjust for delta values equal to 0
        idx_delta_zeros = _np.where(delta==0)[0]
        idx_delta_nozeros = _np.where(delta>0)[0]
        delta[idx_delta_zeros] = _np.min(delta[idx_delta_nozeros])
        
        refractory = 1 / fmax

        maxp, minp, maxv, minv = _PeakDetection(delta=delta, refractory=refractory, start_max=True)(signal)

        if maxp[0] == 0:
            maxp = maxp[1:]

        fsamp = signal.get_sampling_freq()

        ibi_values = _np.diff(maxp) / fsamp
        ibi_values = _np.r_[ibi_values[0], ibi_values]
        idx_ibi = _np.array(maxp)

        ibi = _UnevenlySignal(values=ibi_values,
                              sampling_freq=fsamp,
                              start_time=signal.get_start_time(),
                              signal_type='IBI',
                              x_values=idx_ibi,
                              x_type='indices',
                              duration=signal.get_duration())
        
        return ibi


# PHASIC ESTIMATION
class DriverEstim(_Estimator):
    """
    Estimates the driver of an EDA signal according to (see Notes)

    The estimation uses a deconvolution using a Bateman function as Impulsive Response Function.
    The version of the Bateman function here adopted is:

    :math:`b = e^{-t/T1} - e^{-t/T2}`

    Optional parameters
    -------------------
    t1 : float, >0, default = 0.75
        Value of the T1 parameter of the bateman function
    t2 : float, >0, default = 2
        Value of the T2 parameter of the bateman function

    Returns
    -------
    driver : EvenlySignal
        The EDA driver function

    Notes
    -----
    Please cite:
        
    """
    #TODO: add citation

    def __init__(self, t1=.75, t2=2):
        assert t1 > 0, "t1 value has to be positive"
        assert t2 > 0, "t2 value has to be positive"
        _Estimator.__init__(self, t1=t1, t2=t2)

    @classmethod
    def algorithm(cls, signal, params):
        t1 = params['t1']
        t2 = params['t2']

        fsamp = signal.get_sampling_freq()
        bateman = DriverEstim._gen_bateman(fsamp, [t1, t2])
        idx_max_bat = _np.argmax(bateman)

        # Prepare the input signal to avoid starting/ending peaks in the driver
        bateman_first_half = bateman[0:idx_max_bat + 1]
        bateman_first_half = signal[0] * (bateman_first_half - _np.min(bateman_first_half)) / (
            _np.max(bateman_first_half) - _np.min(bateman_first_half))

        bateman_second_half = bateman[idx_max_bat:]
        bateman_second_half = signal[-1] * (bateman_second_half - _np.min(bateman_second_half)) / (
            _np.max(bateman_second_half) - _np.min(bateman_second_half))

        signal_in = _np.r_[bateman_first_half, signal.get_values(), bateman_second_half]
        signal_in = _EvenlySignal(signal_in, fsamp)

        # deconvolution
        driver = _DeConvolutionalFilter(irf=bateman, normalize=True, deconv_method='fft')(signal_in)
        driver = driver[idx_max_bat + 1: idx_max_bat + len(signal)]

        # gaussian smoothing
        driver = _ConvolutionalFilter(irftype='gauss', win_len=_np.max([0.2, 1 / fsamp]) * 8, normalize=True)(driver)

        driver = _EvenlySignal(driver, sampling_freq=fsamp, start_time=signal.get_start_time(),signal_type="dEDA")
        return driver

    @staticmethod
    def _gen_bateman(fsamp, par_bat):
        """
        Generates the bateman function:

        :math:`b = e^{-t/T1} - e^{-t/T2}`

        Parameters
        ----------
        fsamp : float
            Sampling frequency
        par_bat: list (T1, T2)
            Parameters of the bateman function

        Returns
        -------
        bateman : array
            The bateman function
        """

        idx_T1 = par_bat[0] * fsamp
        idx_T2 = par_bat[1] * fsamp
        len_bat = idx_T2 * 10
        idx_bat = _np.arange(len_bat)
        bateman = _np.exp(-idx_bat / idx_T2) - _np.exp(-idx_bat / idx_T1)

        # normalize
        bateman = fsamp * bateman / _np.sum(bateman)
        return bateman


class PhasicEstim(_Estimator):
    """
    Estimates the phasic and tonic components of a EDA driver function.
    It uses a detection algorithm based on the derivative of the driver.

    
    Parameters:
    -----------
    delta : float, >0
        Minimum amplitude of the peaks in the driver
        
    Optional parameters
    -------------------
    grid_size : float, >0, default = 1
        Sampling size of the interpolation grid
    pre_max : float, >0, default = 2
        Duration (in seconds) of interval before the peak where to search the start of the peak
    post_max : float, >0, default = 2
        Duration (in seconds) of interval after the peak where to search the end of the peak

    Returns:
    --------
    phasic : EvenlySignal
        The phasic component
    tonic : EvenlySignal
        The tonic component
    driver_no_peak : EvenlySignal
        The "de-peaked" driver signal used to generate the interpolation grid
    
    Notes
    -----
    Please cite:
        
    """
    #TODO: add citation

    def __init__(self, delta, grid_size=1, win_pre=2, win_post=2):
        assert delta > 0, "Delta value has to be positive"
        assert grid_size > 0, "Step of the interpolation grid has to be positive"
        assert win_pre > 0,  "Window pre peak value has to be positive"
        assert win_post > 0, "Window post peak value has to be positive"
        _Estimator.__init__(self, delta=delta, grid_size=grid_size, win_pre=win_pre, win_post=win_post)

    @classmethod
    def algorithm(cls, signal, params):
        delta = params["delta"]
        grid_size = params["grid_size"]
        win_pre = params['win_pre']
        win_post = params['win_post']

        fsamp = signal.get_sampling_freq()

        # find peaks in the driver
        idx_max, idx_min, val_max, val_min = _PeakDetection(delta=delta, refractory=1, start_max=True)(signal)

        # identify start and stop of the peak
        idx_pre, idx_post = _PeakSelection(indices=idx_max, win_pre=win_pre, win_post=win_post)(signal)

        # Linear interpolation to substitute the peaks
        driver_no_peak = _np.copy(signal)
        for I in range(len(idx_pre)):
            i_st = idx_pre[I]
            i_sp = idx_post[I]

            if not _np.isnan(i_st) and not _np.isnan(i_sp):
                idx_base = _np.arange(i_sp - i_st)
                coeff = (signal[i_sp] - signal[i_st]) / len(idx_base)
                driver_base = idx_base * coeff + signal[i_st]
                driver_no_peak[i_st:i_sp] = driver_base

        # generate the grid for the interpolation
        idx_grid = _np.arange(0, len(driver_no_peak) - 1, grid_size * fsamp)
        idx_grid = _np.r_[idx_grid, len(driver_no_peak) - 1]

        driver_grid = _UnevenlySignal(driver_no_peak[idx_grid], sampling_freq = fsamp, start_time= signal.get_start_time(), signal_type="dEDA",
                                      x_values=idx_grid, x_type='indices', duration=signal.get_duration())
        tonic = driver_grid.to_evenly(kind='cubic')

        phasic = signal - tonic

        return phasic, tonic, driver_no_peak


class Energy(_Estimator):
    """
    Estimate the local energy of the signal, by windowing

    Parameters
    ----------
    win_len : float, >0
        Length of the window in seconds
    win_step : float, >0
        Shift of the window to start the next window
        
    Optional parameters
    -------------------
    
    smooth : boolean, default = True
        Whether to convolve the result with a gaussian window

    Returns
    -------
    energy : numpy.array
        Local energy
    """

    def __init__(self, win_len, win_step, smooth=True):
        assert win_len > 0, "Window length has to be positive"
        assert win_step > 0, "Window step has to be positive"
        _Estimator.__init__(self, win_len=win_len, win_step=win_step, smooth=smooth)

    @classmethod
    def algorithm(cls, signal, params):
        win_len = params['win_len']
        win_step = params['win_step']
        smooth = params['smooth']

        fsamp = signal.get_sampling_freq()
        idx_len = win_len * fsamp
        idx_step = win_step * fsamp

        windows = _np.arange(0, len(signal) - idx_len + 1, idx_step)

        energy = _np.empty(len(windows) + 2)
        for i in range(1, len(windows) + 1):
            start = windows[i - 1]
            portion_curr = signal.segment_idx(start, start + idx_len)
            energy[i] = _np.nanmean(_np.power(portion_curr, 2))
        energy[0] = energy[1]
        energy[-1] = energy[-2]

        idx_interp = _np.r_[0, windows + round(idx_len / 2), len(signal)-1]
        energy_out = _UnevenlySignal(energy, signal.get_sampling_freq(), start_time = signal.get_start_time(), x_values=idx_interp,
                                     x_type='indices', duration=signal.get_duration()).to_evenly('linear')

        if smooth:
            energy_out = _ConvolutionalFilter(irftype='gauss', win_len=2, normalize=True)(energy_out)

        return energy_out



# coding=utf-8
from abc import abstractmethod as _abstract, ABCMeta as _ABCMeta
from pyphysio.Signal import Signal, EvenlySignal
from pyphysio.Utility import PhUI as _PhUI
import numpy as _np
__author__ = 'AleB'


class Algorithm(object):
    """
    This is the algorithm container super class. It (is abstract) should be used only to be extended.
    """
    __metaclass__ = _ABCMeta

    _log = None

    def __init__(self, **kwargs):
        """
        Incorporates the parameters and saves them in the instance.
        @param params: Dictionary of string-value parameters passed by the user.
        @type params: dict
        @param _kwargs: Internal channel for subclasses kwargs parameters.
        @type _kwargs: dict
        @param kwargs: kwargs parameters to pass to the feature extractor.
        @type kwargs: dict
        """
        self._params = {}
        self.set_unchecked(**kwargs)  # already checked by __init__

    def __call__(self, data):
        """
        Executes the algorithm using the parameters saved by the constructor.
        @param data: The data.
        @type data: TimeSeries
        @return: The result.
        """
        return self.run(data, self._params)

    def __repr__(self):
        return self.__class__.__name__ + str(self._params) if 'name' not in self._params else self._params['name']

    def set_unchecked(self, **kwargs):
        self._params.update(kwargs)

    def set(self, **kwargs):
        kk = self.get()
        kk.update(kwargs)
        self.__init__(**kk)

    def get(self, param=None):
        """
        Placeholder for the subclasses
        @return
        """
        if param is None:
            return self._params
        else:
            return self._params[param]

    @classmethod
    def run(cls, data, params=None, use_cache=False, **kwargs):
        """
        Gets the data from the cache or calculates, caches and returns it.
        @param data: Source data
        @type data: TimeSeries
        @param params: Parameters for the calculator
        @type params: dict
        @param use_cache: Whether to use the cache memory or not
        @type use_cache: bool
        @return: The value of the feature.
        """
        if type(params) is dict:
            kwargs.update(params)
        if not isinstance(data.get_values(), _np.ndarray):
            _PhUI.w("The data must be a Signal (see class EvenlySignal and UnevenlySignal).")
            use_cache = False
        if use_cache is True:
            Cache.cache_check(data)
            # noinspection PyTypeChecker
            return Cache.run_cached(data, cls, kwargs)
        else:            
            if not data.is_multi():
                return cls.algorithm(data, kwargs)
            else:
                data_values = data.get_values()
                values_out = []
                for i_ch in range(data.get_nchannels()):
                    channel_ph = EvenlySignal(data_values[:,i_ch], data.get_sampling_freq(), data.get_start_time())
                    output_ph = cls.algorithm(channel_ph, kwargs)
                    values_out.append(output_ph)
        
                # if output are signals, compose a multimodal instance
                if isinstance(values_out[0], EvenlySignal):
                    values_out_np = _np.stack([x.get_values() for x in values_out], axis=1)
                    output = data.clone_properties(values_out_np)
                    return(output)
                else:
                    return(values_out)

    @classmethod
    @_abstract
    def is_compatible(cls, signal):
        """
        Placeholder for the subclasses
        :returns: Weather nature is compatible or not
        @raise NotImplementedError: Ever
        """
        pass

    @classmethod
    @_abstract
    def algorithm(cls, data, params):
        """
        Placeholder for the subclasses
        @raise NotImplementedError: Ever
        :param params:
        :param data:
        """
        pass

    @classmethod
    def cache_key(cls, params):
        """
        This method computes an hash to use as a part of the key in the cache starting from the parameters used by the
        feature.
        @return: The hash of the parameters used by the feature.
        :param params:
        """
        p = params.copy()
        p.update({'': str(cls)})
        return str(p).replace('\'', '')

    @classmethod
    def log(cls, message):
        l = (_PhUI.i, cls.__name__ + ": " + message)
        cls.emulate_log([l])
        if cls._log is not None:
            cls._log.append(l)

    @classmethod
    def warn(cls, message):
        l = (_PhUI.w, cls.__name__ + ": " + message)
        cls.emulate_log([l])
        if cls._log is not None:
            cls._log.append(l)

    @classmethod
    def error(cls, message, raise_error=False):
        l = (_PhUI.e, cls.__name__ + ": " + message)
        cls.emulate_log([l])
        if cls._log is not None:
            cls._log.append(l)
        assert not raise_error, l

    @classmethod
    def set_logger(cls):
        cls._log = []

    @classmethod
    def unset_logger(cls):
        u = cls._log
        cls._log = None
        return u

    @classmethod
    def emulate_log(cls, log):
        map(lambda f_m: f_m[0](f_m[1]), log)


# noinspection PyProtectedMember
class Cache(object):
    """ Class that gives cache support."""

    def __init__(self):
        pass

    # Field-checked methods

    @staticmethod
    def cache_clear(obj):
        """
        Clears the cache and frees memory (GC?)
        :param obj:
        """
        obj._cache = {}
        obj._mutated = False

    @staticmethod
    def cache_check(obj):
        """
        Checks the presence of the cache structure.
        :param obj:
        """
        if not hasattr(obj, "_cache") or hasattr(obj, "_mutated") and obj._mutated:
            Cache.cache_clear(obj)

    # Field-unchecked methods

    @staticmethod
    def invalidate(obj, algorithm, params):
        """
        Invalidates the specified calculator's cached data if any.
        :type algorithm: Algorithm
        :param obj:
        :param params:
        """
        key = algorithm.cache_key(params)
        if key in obj._cache:
            del obj._cache[key]

    @staticmethod
    def run_cached(obj, algorithm, params):
        """
        Gets data from the cache if valid
        :param params:
        :param obj:
        :type algorithm: Algorithm
        :return: The data or None
        """
        key = algorithm.cache_key(params)

        if key not in obj._cache:
            algorithm.set_logger()
            val = algorithm.algorithm(obj, params)
            log = algorithm.unset_logger()
            obj._cache[key] = (val, log)
        else:
            val, log = obj._cache[key]
            algorithm.emulate_log(log)
        return val



# coding=utf-8
import numpy as _np

from ..BaseIndicator import Indicator as _Indicator
from ..indicators.FrequencyDomain import PowerInBand as _PowerInBand
import scipy.stats as _sps
from ..filters.Filters import ImputeNAN as _ImputeNAN

__author__ = 'AleB'

class Kurtosis(_Indicator):
    """
    Compute the Kurtosis of the signal
    
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        k = _sps.kurtosis(data.get_values())
        return(k)

class Entropy(_Indicator):
    def __init__(self, nbins=25, **kwargs):
        _Indicator.__init__(self, nbins=nbins, **kwargs)
    
    @classmethod
    def algorithm(cls, data, params):
        if _np.isnan(data).all():
            return(_np.nan)
        nbins=params['nbins']
        p_data = _np.histogram(data.get_values(), bins=nbins)[0]/len(data) # calculates the probabilities
        entropy = _sps.entropy(p_data)  # input probabilities to get the entropy 
        return(entropy)

class DerivativeEnergy(_Indicator):
    """
    Compute the Derivative Energy

    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)
    
    @classmethod
    def algorithm(cls, data, params):
        x = data.get_values()
        de = _np.sqrt(_np.nanmean(_np.power(_np.diff(x), 2)))
        return(de)
        
class SpectralPowerRatio(_Indicator):
    """
    Compute the Spectral Power Ratio

    """
    def __init__(self, method='ar', bandN=[5,14], bandD=[5,50],**kwargs):
        _Indicator.__init__(self, method=method, bandN=bandN, bandD=bandD, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        bandN = params['bandN']
        bandD = params['bandD']
        assert bandD[1] < data.get_sampling_freq()/2, 'The higher frequency in bandD is greater than fsamp/2: cannot compute power' # CHECK: check sampling frequency of the signal (e.g. <=128)
        p_N = _PowerInBand(bandN[0], bandN[1], params['method'])(data)
        p_D = _PowerInBand(bandD[0],bandD[1], params['method'])(data)
        return(p_N/p_D)

class CVSignal(_Indicator):
    """
    Compute the Coefficient of variation of the signal
    
    See: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3859838/

    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        mean = _np.nanmean(data)
        sd = _np.nanstd(data)
        cv = sd/mean
        return(cv)

class PercentageNAN(_Indicator):
    """
    Compute the Percentage of NaNs

    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        n_nan = _np.sum(_np.isnan(data))
        data = _ImputeNAN()(data)
        return(100*n_nan/len(data))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  6 14:59:35 2020

@author: bizzego
"""
import numpy as np
import matplotlib.pyplot as plt

from statsmodels.graphics.tsaplots import plot_pacf
from statsmodels.graphics.tsaplots import plot_acf
from scipy.stats import linregress
from statsmodels.tsa.arima_model import ARIMA
from statsmodels.tsa.filters.filtertools import recursive_filter
from statsmodels.tsa.ar_model import AutoReg

LAG = 20

#%%
#x = arima.sim(list(order = c(1,1,0), ar = 0.7), n = 200) #Generates random data from ARIMA(1,1,0). This will generate a new data set for each call.
#z = ts.intersect(x, lag(x,-3), lag(x,-4)) #Creates a matrix z with columns, xt, xt-3, and xt-4
#y = 15+0.8*z[,2]+1.5*z[,3] #Creates y from lags 3 and 4 of randomly generated x
x = np.loadtxt('/home/bizzego/tmp/signal_x.txt', delimiter = ',')
y = np.loadtxt('/home/bizzego/tmp/signal_y.txt', delimiter = ',')

#%%

#ccf(z[,1],y,na.action = na.omit) #CCF between x and y
ccf = np.correlate(x, y, mode='full')

plt.stem(ccf[200-LAG:200+LAG])

#%%
#acf(x)
plot_acf(x, lags=22)

#%%
#ar1model = arima(x, order = c(1,1,0))
model = ARIMA(x, order=(1,1,0))
model_fit = model.fit(disp=0)
print(model_fit.summary())

#%%
#pwx=ar1model$residuals
pwx=model_fit.resid

#%%
newpwy = np.convolve(y, [1, -1.7445, 0.7445], mode='full')

#%%
ccf = np.correlate(pwx, newpwy, mode='full')
plt.plot(ccf)

plot_acf(newpwy)

#%%
model = AutoReg(x, lags=5)
model_fit = model.fit()

coeffs = -model_fit.params[1:]

#%%
pwx = model_fit.resid

plot_acf(pwx)

#%%
newpwy = np.convolve(y, [1]+list(coeffs), mode='full')
plot_acf(newpwy)

#%%
ccf = np.correlate(pwx, newpwy, mode='full')

plt.stem(ccf[200-LAG:200+LAG])

